################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
from orquestra.quantum.api.backend import QuantumBackend
from orquestra.quantum.circuits import Circuit
from orquestra.quantum.measurements import Measurements
from orquestra.quantum.symbolic_simulator import SymbolicSimulator


class MockQuantumBackend(QuantumBackend):

    supports_batching = False

    def __init__(self):
        super().__init__()
        self._simulator = SymbolicSimulator()

    def run_circuit_and_measure(
        self, circuit: Circuit, n_samples: int, **kwargs
    ) -> Measurements:
        super(MockQuantumBackend, self).run_circuit_and_measure(circuit, n_samples)

        return self._simulator.run_circuit_and_measure(circuit, n_samples)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for conversions.py"""
import unittest

import numpy

from orquestra.quantum.openfermion.chem import MolecularData
from orquestra.quantum.openfermion.config import EQ_TOLERANCE
from orquestra.quantum.openfermion.ops.operators import FermionOperator
from orquestra.quantum.openfermion.ops.representations import InteractionOperatorError
from orquestra.quantum.openfermion.transforms.opconversions import get_fermion_operator
from orquestra.quantum.openfermion.transforms.opconversions.term_reordering import (
    normal_ordered,
)
from orquestra.quantum.openfermion.transforms.repconversions.conversions import (
    get_interaction_operator,
    get_molecular_data,
)


class GetInteractionOperatorTest(unittest.TestCase):
    def test_get_molecular_operator(self):
        coefficient = 3.0
        operators = ((2, 1), (3, 0), (0, 0), (3, 1))
        op = FermionOperator(operators, coefficient)

        molecular_operator = get_interaction_operator(op)
        fermion_operator = get_fermion_operator(molecular_operator)
        fermion_operator = normal_ordered(fermion_operator)
        self.assertTrue(normal_ordered(op) == fermion_operator)

        op = FermionOperator("1^ 1")
        op *= 0.5 * EQ_TOLERANCE
        molecular_operator = get_interaction_operator(op)
        self.assertEqual(molecular_operator.constant, 0)
        self.assertTrue(
            numpy.allclose(molecular_operator.one_body_tensor, numpy.zeros((2, 2)))
        )

    def test_get_interaction_operator_bad_input(self):
        with self.assertRaises(TypeError):
            get_interaction_operator("3")

    def test_get_interaction_operator_below_threshold(self):
        op = get_interaction_operator(FermionOperator("1^ 1", 0.0))
        self.assertEqual(op.constant, 0)
        self.assertTrue(numpy.allclose(op.one_body_tensor, numpy.zeros((1, 1))))
        self.assertTrue(numpy.allclose(op.two_body_tensor, numpy.zeros((1, 1, 1, 1))))

    def test_get_interaction_operator_too_few_qubits(self):
        with self.assertRaises(ValueError):
            get_interaction_operator(FermionOperator("3^ 2^ 1 0"), 3)

    def test_get_interaction_operator_bad_1body_term(self):
        with self.assertRaises(InteractionOperatorError):
            get_interaction_operator(FermionOperator("1^ 0^"))

    def test_get_interaction_operator_bad_2body_term(self):
        with self.assertRaises(InteractionOperatorError):
            get_interaction_operator(FermionOperator("3^ 2 1 0"))

    def test_get_interaction_operator_nonmolecular_term(self):
        with self.assertRaises(InteractionOperatorError):
            get_interaction_operator(FermionOperator("3^ 2 1"))

    def test_get_molecular_data(self):
        """Test conversion to MolecularData from InteractionOperator"""
        coefficient = 3.0
        operators = ((2, 1), (3, 0), (0, 0), (3, 1))
        op = FermionOperator(operators, coefficient)

        molecular_operator = get_interaction_operator(op)
        molecule = get_molecular_data(
            molecular_operator,
            geometry=[["H", [0, 0, 0]]],
            basis="aug-cc-pvtz",
            multiplicity=2,
            n_electrons=1,
        )
        self.assertTrue(isinstance(molecule, MolecularData))

        molecule = get_molecular_data(
            molecular_operator,
            geometry=[["H", [0, 0, 0]]],
            basis="aug-cc-pvtz",
            multiplicity=2,
            n_electrons=1,
            reduce_spin=False,
        )
        self.assertTrue(isinstance(molecule, MolecularData))



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for conversions.py"""
import unittest

import pytest
import sympy

from orquestra.quantum.openfermion import check_no_sympy, get_fermion_operator
from orquestra.quantum.openfermion.ops.operators import FermionOperator, QubitOperator


def test_get_fermion_operator_wrong_type():
    with pytest.raises(TypeError):
        _ = get_fermion_operator(QubitOperator())


class RaisesSympyExceptionTest(unittest.TestCase):
    def test_raises_sympy_expression(self):
        operator = FermionOperator("0^", sympy.Symbol("x"))
        with self.assertRaises(TypeError):
            check_no_sympy(operator)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This file contains tests of code performance to reveal bottlenecks."""
import logging
import time

import numpy

from orquestra.quantum.openfermion.hamiltonians import jellium_model
from orquestra.quantum.openfermion.linalg import (
    LinearQubitOperator,
    LinearQubitOperatorOptions,
    ParallelLinearQubitOperator,
    jordan_wigner_sparse,
)
from orquestra.quantum.openfermion.ops import FermionOperator, QubitOperator
from orquestra.quantum.openfermion.testing.testing_utils import (
    random_interaction_operator,
)
from orquestra.quantum.openfermion.transforms import (
    commutator_ordered_diagonal_coulomb_with_two_body_operator,
)
from orquestra.quantum.openfermion.transforms.opconversions import (
    get_fermion_operator,
    jordan_wigner,
    normal_ordered,
)
from orquestra.quantum.openfermion.utils import Grid, commutator


def benchmark_molecular_operator_jordan_wigner(n_qubits):
    """Test speed with which molecular operators transform to qubit operators.

    Args:
        n_qubits: The size of the molecular operator instance. Ideally, we
            would be able to transform to a qubit operator for 50 qubit
            instances in less than a minute. We are way too slow right now.

    Returns:
        runtime: The number of seconds required to make the conversion.
    """
    # Get an instance of InteractionOperator.
    molecular_operator = random_interaction_operator(n_qubits)

    # Convert to a qubit operator.
    start = time.time()
    _ = jordan_wigner(molecular_operator)
    end = time.time()

    # Return runtime.
    runtime = end - start
    return runtime


def benchmark_fermion_math_and_normal_order(n_qubits, term_length, power):
    """Benchmark both arithmetic operators and normal ordering on fermions.

    The idea is we generate two random FermionTerms, A and B, each acting
    on n_qubits with term_length operators. We then compute
    (A + B) ** power. This is costly that is the first benchmark. The second
    benchmark is in normal ordering whatever comes out.

    Args:
        n_qubits: The number of qubits on which these terms act.
        term_length: The number of operators in each term.
        power: Int, the exponent to which to raise sum of the two terms.

    Returns:
        runtime_math: The time it takes to perform (A + B) ** power
        runtime_normal_order: The time it takes to perform
            FermionOperator.normal_order()
    """
    # Generate random operator strings.
    operators_a = [(numpy.random.randint(n_qubits), numpy.random.randint(2))]
    operators_b = [(numpy.random.randint(n_qubits), numpy.random.randint(2))]
    for _ in range(term_length):

        # Make sure the operator is not trivially zero.
        operator_a = (numpy.random.randint(n_qubits), numpy.random.randint(2))
        while operator_a == operators_a[-1]:
            operator_a = (numpy.random.randint(n_qubits), numpy.random.randint(2))
        operators_a += [operator_a]

        # Do the same for the other operator.
        operator_b = (numpy.random.randint(n_qubits), numpy.random.randint(2))
        while operator_b == operators_b[-1]:
            operator_b = (numpy.random.randint(n_qubits), numpy.random.randint(2))
        operators_b += [operator_b]

    # Initialize FermionTerms and then sum them together.
    fermion_term_a = FermionOperator(tuple(operators_a), float(numpy.random.randn()))
    fermion_term_b = FermionOperator(tuple(operators_b), float(numpy.random.randn()))
    fermion_operator = fermion_term_a + fermion_term_b

    # Exponentiate.
    start_time = time.time()
    fermion_operator **= power
    runtime_math = time.time() - start_time

    # Normal order.
    start_time = time.time()
    normal_ordered(fermion_operator)
    runtime_normal_order = time.time() - start_time

    # Return.
    return runtime_math, runtime_normal_order


def benchmark_jordan_wigner_sparse(n_qubits):
    """Benchmark the speed at which a FermionOperator is mapped to a matrix.

    Args:
        n_qubits: The number of qubits in the example.

    Returns:
        runtime: The time in seconds that the benchmark took.
    """
    # Initialize a random FermionOperator.
    molecular_operator = random_interaction_operator(n_qubits)
    fermion_operator = get_fermion_operator(molecular_operator)

    # Map to SparseOperator class.
    start_time = time.time()
    _ = jordan_wigner_sparse(fermion_operator)
    runtime = time.time() - start_time
    return runtime


def benchmark_linear_qubit_operator(n_qubits, n_terms, processes=None):
    """Test speed with getting a linear operator from a Qubit Operator.

    Args:
        n_qubits: The number of qubits, implying the dimension of the operator
            is 2 ** n_qubits.
        n_terms: The number of terms in a qubit operator.
        processes: The number of processors to use.

    Returns:
        runtime_operator: The time it takes to get the linear operator.
        runtime_matvec: The time it takes to perform matrix multiplication.
    """
    # Generates Qubit Operator with specified number of terms.
    map_int_to_operator = {
        0: "X",
        1: "Y",
        2: "Z",
    }
    qubit_operator = QubitOperator.zero()
    for _ in range(n_terms):
        tuples = []
        for i in range(n_qubits):
            operator = numpy.random.randint(4)
            # 3 is 'I', so just skip.
            if operator > 2:
                continue
            tuples.append((i, map_int_to_operator[operator]))
        if tuples:
            qubit_operator += QubitOperator(tuples, 1.00)

    # Gets an instance of (Parallel)LinearQubitOperator.
    start = time.time()
    if processes is None:
        linear_operator = LinearQubitOperator(qubit_operator, n_qubits)
    else:
        linear_operator = ParallelLinearQubitOperator(
            qubit_operator, n_qubits, LinearQubitOperatorOptions(processes=processes)
        )

    end = time.time()
    runtime_operator = end - start

    vec = numpy.random.rand(2 ** n_qubits)
    # Performs matrix multiplication.
    start = time.time()
    _ = linear_operator * vec
    end = time.time()
    runtime_matvec = end - start
    return runtime_operator, runtime_matvec


def benchmark_commutator_diagonal_coulomb_operators_2D_spinless_jellium(side_length):
    """Test speed of computing commutators using specialized functions.

    Args:
        side_length: The side length of the 2D jellium grid. There are
            side_length ** 2 qubits, and O(side_length ** 4) terms in the
            Hamiltonian.

    Returns:
        runtime_commutator: The time it takes to compute a commutator, after
            partitioning the terms and normal ordering, using the regular
            commutator function.
        runtime_diagonal_commutator: The time it takes to compute the same
            commutator using methods restricted to diagonal Coulomb operators.
    """
    hamiltonian = normal_ordered(
        jellium_model(Grid(2, side_length, 1.0), plane_wave=False)
    )

    part_a = FermionOperator.zero()
    part_b = FermionOperator.zero()
    add_to_a_or_b = 0  # add to a if 0; add to b if 1
    for term, coeff in hamiltonian.terms.items():
        # Partition terms in the Hamiltonian into part_a or part_b
        if add_to_a_or_b:
            part_a += FermionOperator(term, coeff)
        else:
            part_b += FermionOperator(term, coeff)
        add_to_a_or_b ^= 1

    start = time.time()
    _ = normal_ordered(commutator(part_a, part_b))
    end = time.time()
    runtime_commutator = end - start

    start = time.time()
    _ = commutator_ordered_diagonal_coulomb_with_two_body_operator(part_a, part_b)
    end = time.time()
    runtime_diagonal_commutator = end - start

    return runtime_commutator, runtime_diagonal_commutator


# Sets up each benchmark run.
def run_molecular_operator_jordan_wigner(n_qubits=18):
    """Run InteractionOperator.jordan_wigner_transform() benchmark."""
    logging.info("Starting test on " "InteractionOperator.jordan_wigner_transform()")
    logging.info("n_qubits = %d.", n_qubits)
    runtime = benchmark_molecular_operator_jordan_wigner(n_qubits)
    logging.info(
        "InteractionOperator.jordan_wigner_transform() takes %f " "seconds.\n", runtime
    )

    return runtime


def run_fermion_math_and_normal_order(n_qubits=20, term_length=10, power=15):
    """Run benchmark on FermionOperator math and normal-ordering."""
    logging.info("Starting test on FermionOperator math and normal ordering.")
    logging.info(
        "(n_qubits, term_length, power) = (%d, %d, %d).", n_qubits, term_length, power
    )
    runtime_math, runtime_normal = benchmark_fermion_math_and_normal_order(
        n_qubits, term_length, power
    )
    logging.info(
        "Math took %f seconds. Normal ordering took %f seconds.\n",
        runtime_math,
        runtime_normal,
    )

    return runtime_math, runtime_normal


def run_jordan_wigner_sparse(n_qubits=10):
    """Run FermionOperator.jordan_wigner_sparse() benchmark."""
    logging.info("Starting test on FermionOperator.jordan_wigner_sparse().")
    logging.info("n_qubits = %d.", n_qubits)
    runtime = benchmark_jordan_wigner_sparse(n_qubits)
    logging.info("Construction of SparseOperator took %f seconds.\n", runtime)

    return runtime


def run_linear_qubit_operator(n_qubits=16, n_terms=10, processes=10):
    """Run linear_qubit_operator benchmark."""

    logging.info("Starting test on linear_qubit_operator().")
    logging.info("(n_qubits, n_terms) = (%d, %d).", n_qubits, n_terms)
    _, runtime_sequential = benchmark_linear_qubit_operator(n_qubits, n_terms)
    _, runtime_parallel = benchmark_linear_qubit_operator(n_qubits, n_terms, processes)
    logging.info(
        "LinearQubitOperator took %f seconds, while "
        "ParallelQubitOperator took %f seconds with %d processes, "
        "and ratio is %.2f.\n",
        runtime_sequential,
        runtime_parallel,
        processes,
        runtime_sequential / runtime_parallel,
    )

    return runtime_sequential, runtime_parallel


def run_diagonal_commutator(side_length=4):
    """Run commutator_diagonal_coulomb_operators benchmark."""

    logging.info(
        "Starting test on "
        "commutator_ordered_diagonal_coulomb_with_two_body_operator()."
    )
    (
        runtime_commutator,
        runtime_diagonal_commutator,
    ) = benchmark_commutator_diagonal_coulomb_operators_2D_spinless_jellium(
        side_length=side_length
    )
    logging.info(
        "Regular commutator computation took %f seconds, while "
        "commutator_ordered_diagonal_coulomb_with_two_body_operator"
        " took %f seconds. Ratio is %.2f.\n",
        runtime_commutator,
        runtime_diagonal_commutator,
        runtime_commutator / runtime_diagonal_commutator,
    )

    return runtime_commutator, runtime_diagonal_commutator



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import itertools
import random
from typing import Iterable, List, Optional

import numpy as np

from orquestra.quantum.circuits import Circuit, X, Y, Z
from orquestra.quantum.measurements import ExpectationValues
from orquestra.quantum.openfermion import (
    FermionOperator,
    InteractionOperator,
    InteractionRDM,
    PolynomialTensor,
    QubitOperator,
    count_qubits,
)
from orquestra.quantum.openfermion import expectation as openfermion_expectation
from orquestra.quantum.openfermion import (
    get_interaction_operator,
    get_sparse_operator,
    normal_ordered,
)
from orquestra.quantum.openfermion.hamiltonians.special_operators import number_operator
from orquestra.quantum.openfermion.linalg import jw_get_ground_state_at_particle_number
from orquestra.quantum.openfermion.transforms import (
    freeze_orbitals,
    get_fermion_operator,
)
from orquestra.quantum.utils import ValueEstimate, bin2dec, dec2bin
from orquestra.quantum.wavefunction import Wavefunction


def get_qubitop_from_matrix(operator: List[List]) -> QubitOperator:
    r"""Expands a 2^n by 2^n matrix into n-qubit Pauli basis. The runtime of
    this function is O(2^2n).

    Args:
        operator: a list of lists (rows) representing a 2^n by 2^n
            matrix.

    Returns:
        A QubitOperator instance corresponding to the expansion of
        the input operator as a sum of Pauli strings:

        O = 2^-n \sum_P tr(O*P) P
    """

    nrows = len(operator)
    ncols = len(operator[0])

    # Check if the input operator is square
    if nrows != ncols:
        raise Exception("The input operator is not square")

    # Check if the dimensions are powers of 2
    if not (((nrows & (nrows - 1)) == 0) and nrows > 0):
        raise Exception("The number of rows is not a power of 2")
    if not (((ncols & (ncols - 1)) == 0) and ncols > 0):
        raise Exception("The number of cols is not a power of 2")

    n = int(np.log2(nrows))  # number of qubits

    def decode(bit_string):  # Helper function for converting any 2n-bit
        # string to a label vector representing a Pauli
        # string of length n

        if len(bit_string) != 2 * n:
            raise Exception("LH_expand:decode: input bit string length not 2n")

        output_label = list(np.zeros(n))
        for i in range(0, n):
            output_label[i] = bin2dec(bit_string[2 * i : 2 * i + 2])

        return output_label

    def trace_product(label_vec):  # Helper function for computing tr(OP)
        # where O is the input operator and P is a
        # Pauli string operator

        def f(j):  # Function which computes the index of the nonzero
            # element in P for a given column j

            j_str = dec2bin(j, n)
            for index in range(0, n):
                if label_vec[index] in [1, 2]:  # flip if X or Y
                    j_str[index] = int(not j_str[index])
            return bin2dec(j_str)

        def nz(j):  # Function which computes the value of the nonzero
            # element in P on the column j

            val_nz = 1.0
            j_str = dec2bin(j, n)
            for index in range(0, n):
                if label_vec[index] == 2:
                    if j_str[index] == 0:
                        val_nz = val_nz * (1j)
                    if j_str[index] == 1:
                        val_nz = val_nz * (-1j)
                if label_vec[index] == 3:
                    if j_str[index] == 1:
                        val_nz = val_nz * (-1)
            return val_nz

        # Compute the trace
        tr = 0.0
        for j in range(0, 2 ** n):  # loop over the columns
            tr = tr + operator[j][f(j)] * nz(j)

        return tr / 2 ** n

    # Expand the operator in Pauli basis
    coeffs = list(np.zeros(4 ** n))
    labels = list(np.zeros(4 ** n))
    for i in range(0, 4 ** n):  # loop over all 2n-bit strings
        current_string = dec2bin(i, 2 * n)  # see util.py
        current_label = decode(current_string)
        coeffs[i] = trace_product(current_label)
        labels[i] = current_label

    return get_qubitop_from_coeffs_and_labels(coeffs, labels)


def get_qubitop_from_coeffs_and_labels(
    coeffs: List[float], labels: List[List[int]]
) -> QubitOperator:
    """Generates a QubitOperator based on a coefficient vector and
    a label matrix.

    Args:
        coeffs: a list of floats representing the coefficients
            for the terms in the Hamiltonian
        labels: a list of lists (a matrix) where each list
            is a vector of integers representing the Pauli
            string. See pauliutil.py for details.

    Example:

        The Hamiltonian H = 0.1 X1 X2 - 0.4 Y1 Y2 Z3 Z4 can be
        initiated by calling

        H = QubitOperator([0.1, -0.4],  # coefficients
            [[1 1 0 0],  # label matrix
            [2 2 3 3]])
    """

    output = QubitOperator()
    for i in range(0, len(labels)):
        string_term = ""
        for ind, elem in enumerate(labels[i]):
            pauli_symbol = ""
            if elem == 1:
                pauli_symbol = "X" + str(ind) + " "
            if elem == 2:
                pauli_symbol = "Y" + str(ind) + " "
            if elem == 3:
                pauli_symbol = "Z" + str(ind) + " "
            string_term += pauli_symbol

        output += coeffs[i] * QubitOperator(string_term)

    return output


def generate_random_qubitop(
    nqubits: int,
    nterms: int,
    nlocality: int,
    max_coeff: float,
    fixed_coeff: bool = False,
) -> QubitOperator:
    """Generates a Hamiltonian with term coefficients uniformly distributed
    in [-max_coeff, max_coeff].

    Args:
        nqubits - number of qubits
        nterms    - number of terms in the Hamiltonian
        nlocality - locality of the Hamiltonian
        max_coeff - bound for generating the term coefficients
        fixed_coeff (bool) - If true, all the terms are assign the
            max_coeff as coefficient.

    Returns:
        A QubitOperator with the appropriate coefficient vector
        and label matrix.
    """
    # generate random coefficient vector
    if fixed_coeff:
        coeffs = [max_coeff] * nterms
    else:
        coeffs = list(np.zeros(nterms))
        for j in range(0, nterms):
            coeffs[j] = random.uniform(-max_coeff, max_coeff)

    # generate random label vector
    labels = list(np.zeros(nterms, dtype=int))
    label_set = set()
    j = 0
    while j < nterms:
        inds_nontrivial = sorted(random.sample(range(0, nqubits), nlocality))
        label = list(np.zeros(nqubits, dtype=int))
        for ind in inds_nontrivial:
            label[ind] = random.randint(1, 3)
        if str(label) not in label_set:
            labels[j] = label
            j += 1
            label_set.add(str(label))
    return get_qubitop_from_coeffs_and_labels(coeffs, labels)


def evaluate_qubit_operator(
    qubit_operator: QubitOperator, expectation_values: ExpectationValues
) -> ValueEstimate:
    """Evaluate the expectation value of a qubit operator using expectation values for
    the terms.

    Args:
        qubit_operator: the operator
        expectation_values: the expectation values

    Returns:
        value_estimate: stores the value of the expectation and its precision
    """

    # Sum the contributions from all terms
    total = 0

    # Add all non-trivial terms
    term_index = 0
    for term in qubit_operator.terms:
        total += np.real(
            qubit_operator.terms[term] * expectation_values.values[term_index]
        )
        term_index += 1

    value_estimate = ValueEstimate(total)
    return value_estimate


def evaluate_qubit_operator_list(
    qubit_operator_list: List[QubitOperator], expectation_values: ExpectationValues
) -> ValueEstimate:
    """Evaluate the expectation value of a qubit operator list using expectation values
    for the terms. The expectation values should be in the order given by the qubit
    operator list, and the value returned is the sum of all terms in the qubit operator
    list.

    Args:
        qubit_operator_list: the operator list
        expectation_values: the expectation values

    Returns:
        value_estimate: stores the value of the expectation and its precision
    """

    # Sum the contributions from all terms
    total = 0

    # Add all non-trivial terms
    term_index = 0
    for qubit_operator in qubit_operator_list:
        for term in qubit_operator.terms:
            total += np.real(
                qubit_operator.terms[term] * expectation_values.values[term_index]
            )
            term_index += 1

    value_estimate = ValueEstimate(total)
    return value_estimate


def reverse_qubit_order(qubit_operator: QubitOperator, n_qubits: Optional[int] = None):
    """Reverse the order of qubit indices in a qubit operator.

    Args:
        qubit_operator (openfermion.QubitOperator): the operator
        n_qubits (int): total number of qubits. Needs to be provided when
            the size of the system of interest is greater than the size of qubit
            operator (optional)

    Returns:
        reversed_op (openfermion.ops.QubitOperator)
    """

    reversed_op = QubitOperator()

    if n_qubits is None:
        n_qubits = count_qubits(qubit_operator)
    if n_qubits < count_qubits(qubit_operator):
        raise ValueError("Invalid number of qubits specified.")

    for term in qubit_operator.terms:
        new_term = []
        for factor in term:
            new_factor = list(factor)
            new_factor[0] = n_qubits - 1 - new_factor[0]
            new_term.append(tuple(new_factor))
        reversed_op += QubitOperator(tuple(new_term), qubit_operator.terms[term])
    return reversed_op


def get_expectation_value(
    qubit_op: QubitOperator, wavefunction: Wavefunction, reverse_operator: bool = False
) -> complex:
    """Get the expectation value of a qubit operator with respect to a wavefunction.
    Args:
        qubit_op: the operator
        wavefunction: the wavefunction
        reverse_operator: whether to reverse order of qubit operator
            before computing expectation value. This should be True if the convention
            of the basis states used for the wavefunction is the opposite of the one in
            the qubit operator. This is the case when the wavefunction uses
            Rigetti convention (https://arxiv.org/abs/1711.02086) of ordering qubits.
    Returns:
        the expectation value
    """
    n_qubits = wavefunction.amplitudes.shape[0].bit_length() - 1

    # Convert the qubit operator to a sparse matrix. Note that the qubit indices
    # must be reversed because OpenFermion and our Wavefunction use
    # different conventions for how to order the computational basis states!
    if reverse_operator:
        qubit_op = reverse_qubit_order(qubit_op, n_qubits=n_qubits)
    sparse_op = get_sparse_operator(qubit_op, n_qubits=n_qubits)

    # Computer the expectation value
    exp_val = openfermion_expectation(sparse_op, wavefunction.amplitudes)
    return exp_val


def change_operator_type(operator, operatorType):
    """Take an operator and attempt to cast it to an operator of a different type

    Args:
        operator: The operator
        operatorType: The type of the operator that the original operator is
            cast to
    Returns:
        An operator with type operatorType
    """
    new_operator = operatorType()
    for op in operator.terms:
        new_operator += operatorType(tuple(op), operator.terms[op])

    return new_operator


def get_fermion_number_operator(n_qubits, n_particles=None):
    """Return a FermionOperator representing the number operator for n qubits.
    If `n_particles` is specified, it can be used for creating constraint on the number
    of particles.

    Args:
        n_qubits (int): number of qubits in the system
        n_particles (int): number of particles in the system.
            If specified, it is subtracted from the number
            operator such as expectation value is zero.
    Returns:
         (openfermion.ops.FermionOperator): the number operator
    """
    operator = number_operator(n_qubits)
    if n_particles is not None:
        operator += FermionOperator("", -1.0 * float(n_particles))
    return get_interaction_operator(operator)


def get_diagonal_component(operator):
    if isinstance(operator, InteractionOperator):
        return _get_diagonal_component_interaction_operator(operator)
    elif isinstance(operator, PolynomialTensor):
        return _get_diagonal_component_polynomial_tensor(operator)
    else:
        raise TypeError(
            f"Getting diagonal component not supported for {0}".format(type(operator))
        )


def _get_diagonal_component_polynomial_tensor(polynomial_tensor):
    """Get the component of an interaction operator that is
    diagonal in the computational basis under Jordan-Wigner
    transformation (i.e., the terms that can be expressed
    as products of number operators).
    Args:
        interaction_operator (openfermion.ops.InteractionOperator): the operator

    Returns:
        tuple: two openfermion.ops.InteractionOperator objects. The first is the
            diagonal component, and the second is the remainder.
    """
    n_modes = count_qubits(polynomial_tensor)
    remainder_tensors = {}
    diagonal_tensors = {}

    diagonal_tensors[()] = polynomial_tensor.constant
    for key in polynomial_tensor.n_body_tensors:
        if key == ():
            continue
        remainder_tensors[key] = np.zeros((n_modes,) * len(key), complex)
        diagonal_tensors[key] = np.zeros((n_modes,) * len(key), complex)

        for indices in itertools.product(range(n_modes), repeat=len(key)):
            creation_counts = {}
            annihilation_counts = {}

            for meta_index, index in enumerate(indices):
                if key[meta_index] == 0:
                    if annihilation_counts.get(index) is None:
                        annihilation_counts[index] = 1
                    else:
                        annihilation_counts[index] += 1
                elif key[meta_index] == 1:
                    if creation_counts.get(index) is None:
                        creation_counts[index] = 1
                    else:
                        creation_counts[index] += 1

            term_is_diagonal = True
            for index in creation_counts:
                if creation_counts[index] != annihilation_counts.get(index):
                    term_is_diagonal = False
                    break
            if term_is_diagonal:
                for index in annihilation_counts:
                    if annihilation_counts[index] != creation_counts.get(index):
                        term_is_diagonal = False
                        break
            if term_is_diagonal:
                diagonal_tensors[key][indices] = polynomial_tensor.n_body_tensors[key][
                    indices
                ]
            else:
                remainder_tensors[key][indices] = polynomial_tensor.n_body_tensors[key][
                    indices
                ]

    return PolynomialTensor(diagonal_tensors), PolynomialTensor(remainder_tensors)


def _get_diagonal_component_interaction_operator(interaction_operator):
    """Get the component of an interaction operator that is
    diagonal in the computational basis under Jordan-Wigner
    transformation (i.e., the terms that can be expressed
    as products of number operators).
    Args:
        interaction_operator (openfermion.ops.InteractionOperator): the operator

    Returns:
        tuple: two openfermion.ops.InteractionOperator objects. The first is the
            diagonal component, and the second is the remainder.
    """

    one_body_tensor = np.zeros(
        interaction_operator.one_body_tensor.shape, dtype=complex
    )
    two_body_tensor = np.zeros(
        interaction_operator.two_body_tensor.shape, dtype=complex
    )
    diagonal_op = InteractionOperator(
        interaction_operator.constant, one_body_tensor, two_body_tensor
    )

    one_body_tensor = np.copy(interaction_operator.one_body_tensor).astype(complex)
    two_body_tensor = np.copy(interaction_operator.two_body_tensor).astype(complex)
    remainder_op = InteractionOperator(0.0, one_body_tensor, two_body_tensor)

    n_spin_orbitals = interaction_operator.two_body_tensor.shape[0]

    for p in range(n_spin_orbitals):
        for q in range(n_spin_orbitals):
            diagonal_op.two_body_tensor[
                p, q, p, q
            ] = interaction_operator.two_body_tensor[p, q, p, q]
            diagonal_op.two_body_tensor[
                p, q, q, p
            ] = interaction_operator.two_body_tensor[p, q, q, p]
            remainder_op.two_body_tensor[p, q, p, q] = 0.0
            remainder_op.two_body_tensor[p, q, q, p] = 0.0

    for p in range(n_spin_orbitals):
        diagonal_op.one_body_tensor[p, p] = interaction_operator.one_body_tensor[p, p]
        remainder_op.one_body_tensor[p, p] = 0.0

    return diagonal_op, remainder_op


def get_polynomial_tensor(fermion_operator, n_qubits=None):
    r"""Convert a fermionic operator to a Polynomial Tensor.

    Args:
        fermion_operator (openferion.ops.FermionOperator): The operator.
        n_qubits (int): The number of qubits to be included in the
            PolynomialTensor. Must be at least equal to the number of qubits
            that are acted on by fermion_operator. If None, then the number of
            qubits is inferred from fermion_operator.

    Returns:
        openfermion.ops.PolynomialTensor: The tensor representation of the
            operator.
    """
    if not isinstance(fermion_operator, FermionOperator):
        raise TypeError("Input must be a FermionOperator.")

    if n_qubits is None:
        n_qubits = count_qubits(fermion_operator)
    if n_qubits < count_qubits(fermion_operator):
        raise ValueError("Invalid number of qubits specified.")

    # Normal order the terms and initialize.
    fermion_operator = normal_ordered(fermion_operator)
    tensor_dict = {}

    # Loop through terms and assign to matrix.
    for term in fermion_operator.terms:
        coefficient = fermion_operator.terms[term]

        # Handle constant shift.
        if len(term) == 0:
            tensor_dict[()] = coefficient

        else:
            key = tuple([operator[1] for operator in term])
            if tensor_dict.get(key) is None:
                tensor_dict[key] = np.zeros((n_qubits,) * len(key), complex)

            indices = tuple([operator[0] for operator in term])
            tensor_dict[key][indices] = coefficient

    return PolynomialTensor(tensor_dict)


def create_circuits_from_qubit_operator(qubit_operator: QubitOperator) -> List[Circuit]:
    """Creates a list of circuit objects from the Pauli terms of a QubitOperator
    Args:
        qubit_operator: operator for which the Pauli terms are converted into circuits

    Return:
        circuit_set: a list of Pauli string gate circuits
    """

    # Get the Pauli terms, ignoring coefficients
    pauli_terms = list(qubit_operator.terms.keys())
    term_gate_map = {"X": X, "Y": Y, "Z": Z}
    circuit_set = []

    # Loop over Pauli terms and populate circuit set list
    for term in pauli_terms:

        circuit = Circuit()

        # Loop over Pauli factors in Pauli term and construct Pauli term circuit
        for pauli in term:  # loop over pauli operators in an n qubit pauli term
            pauli_index = pauli[0]
            pauli_factor = pauli[1]
            circuit += term_gate_map[pauli_factor](pauli_index)

        circuit_set += [circuit]

    return circuit_set


def get_ground_state_rdm_from_qubit_op(
    qubit_operator: QubitOperator, n_particles: int
) -> InteractionRDM:
    """Diagonalize operator and compute the ground state 1- and 2-RDM

    Args:
        qubit_operator: The openfermion operator to diagonalize
        n_particles: number of particles in the target ground state

    Returns:
        rdm: interaction RDM of the ground state with the particle number n_particles
    """

    sparse_operator = get_sparse_operator(qubit_operator)
    e, ground_state_wf = jw_get_ground_state_at_particle_number(
        sparse_operator, n_particles
    )  # float/np.array pair
    n_qubits = count_qubits(qubit_operator)

    one_body_tensor_list = []
    for i in range(n_qubits):
        for j in range(n_qubits):
            idag_j = get_sparse_operator(
                FermionOperator(f"{i}^ {j}"), n_qubits=n_qubits
            )
            idag_j = idag_j.toarray()
            one_body_tensor_list.append(
                np.conjugate(ground_state_wf) @ idag_j @ ground_state_wf
            )

    one_body_tensor = np.array(one_body_tensor_list)
    one_body_tensor = one_body_tensor.reshape(n_qubits, n_qubits)

    two_body_tensor = np.zeros((n_qubits,) * 4, dtype=complex)
    for p in range(n_qubits):
        for q in range(0, p + 1):
            for r in range(n_qubits):
                for s in range(0, r + 1):
                    pdag_qdag_r_s = get_sparse_operator(
                        FermionOperator(f"{p}^ {q}^ {r} {s}"), n_qubits=n_qubits
                    )
                    pdag_qdag_r_s = pdag_qdag_r_s.toarray()
                    rdm_element = (
                        np.conjugate(ground_state_wf) @ pdag_qdag_r_s @ ground_state_wf
                    )
                    two_body_tensor[p, q, r, s] = rdm_element
                    two_body_tensor[q, p, r, s] = -rdm_element
                    two_body_tensor[q, p, s, r] = rdm_element
                    two_body_tensor[p, q, s, r] = -rdm_element

    return InteractionRDM(one_body_tensor, two_body_tensor)


def remove_inactive_orbitals(
    interaction_op: InteractionOperator, n_active: int = None, n_core: int = 0
) -> InteractionOperator:
    """Remove orbitals not in the active space from an interaction operator.

    Args:
        interaction_op: the operator, assumed to be ordered with alternating spin-up and
            spin-down spin orbitals.
        n_active: the number of active molecular orbitals. If None, include all orbitals
            beyond n_core. Note that the number of active spin orbitals will be twice
            the number of active molecular orbitals.
        n_core: the number of core molecular orbitals to be frozen.

    Returns:
        The interaction operator with inactive orbitals removed, and the Hartree-Fock
            energy of the core orbitals added to the constant.
    """

    # This implementation is probably not very efficient, because it converts the
    # interaction operator into a fermion operator and then back to an interaction
    # operator.

    # Convert the InteractionOperator to a FermionOperator
    fermion_op = get_fermion_operator(interaction_op)

    # Determine which occupied spin-orbitals are to be frozen
    occupied = range(2 * n_core)

    unoccupied: Iterable
    # Determine which unoccupied spin-orbitals are to be frozen
    if n_active is not None:
        unoccupied = range(
            2 * n_core + 2 * n_active, interaction_op.one_body_tensor.shape[0]
        )
    else:
        unoccupied = []

    # Freeze the spin-orbitals
    frozen_fermion_op = freeze_orbitals(fermion_op, occupied, unoccupied)

    # Convert back to an interaction operator
    frozen_interaction_op = get_interaction_operator(frozen_fermion_op)

    return frozen_interaction_op


def hf_rdm(n_alpha: int, n_beta: int, n_orbitals: int) -> InteractionRDM:
    """Construct the RDM corresponding to a Hartree-Fock state.

    Args:
        n_alpha (int): number of spin-up electrons
        n_beta (int): number of spin-down electrons
        n_orbitals (int): number of spatial orbitals (not spin orbitals)

    Returns:
        openfermion.ops.InteractionRDM: the reduced density matrix
    """
    # Determine occupancy of each spin orbital
    occ = np.zeros(2 * n_orbitals)
    occ[: (2 * n_alpha) : 2] = 1
    occ[1 : (2 * n_beta + 1) : 2] = 1

    one_body_tensor = np.diag(occ)

    two_body_tensor = np.zeros([2 * n_orbitals for i in range(4)])
    for i in range(2 * n_orbitals):
        for j in range(2 * n_orbitals):
            if i != j and occ[i] and occ[j]:
                two_body_tensor[i, j, j, i] = 1
                two_body_tensor[i, j, i, j] = -1

    return InteractionRDM(one_body_tensor, two_body_tensor)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module constructs Hamiltonians for the multiband Fermi-Hubbard model.
"""

from collections import namedtuple

from orquestra.quantum.openfermion.ops.operators import FermionOperator

TunnelingParameter = namedtuple(
    "TunnelingParameter", ("edge_type", "dofs", "coefficient")
)
InteractionParameter = namedtuple(
    "InteractionParameter", ("edge_type", "dofs", "coefficient", "spin_pairs")
)
PotentialParameter = namedtuple("PotentialParameter", ("dof", "coefficient"))


def number_operator(i, coefficient=1.0, particle_hole_symmetry=False):
    op = FermionOperator(((i, 1), (i, 0)), coefficient)
    if particle_hole_symmetry:
        op -= FermionOperator((), 0.5)
    return op


def interaction_operator(i, j, coefficient=1.0, particle_hole_symmetry=False):
    return number_operator(
        i, coefficient, particle_hole_symmetry=particle_hole_symmetry
    ) * number_operator(j, particle_hole_symmetry=particle_hole_symmetry)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
# isort: skip_file

from .commutators import anticommutator, commutator, double_commutator
from .indexing import down_index, up_index, up_then_down
from .operator_utils import (
    count_qubits,
    get_file_path,
    hermitian_conjugated,
    is_hermitian,
    is_identity,
    load_operator,
    save_operator,
    OperatorSpecificationError,
    OperatorUtilsError,
)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from ._io import (  # noqa: F403
    convert_dict_to_interaction_op,
    convert_dict_to_interaction_rdm,
    convert_dict_to_isingop,
    convert_dict_to_operator,
    convert_dict_to_qubitop,
    convert_interaction_op_to_dict,
    convert_interaction_rdm_to_dict,
    convert_isingop_to_dict,
    convert_qubitop_to_dict,
    get_pauli_strings,
    load_interaction_operator,
    load_interaction_rdm,
    load_ising_operator,
    load_qubit_operator,
    load_qubit_operator_set,
    save_interaction_operator,
    save_interaction_rdm,
    save_ising_operator,
    save_parameter_grid_evaluation,
    save_qubit_operator,
    save_qubit_operator_set,
)
from ._utils import (  # noqa: F403
    change_operator_type,
    create_circuits_from_qubit_operator,
    evaluate_qubit_operator,
    evaluate_qubit_operator_list,
    generate_random_qubitop,
    get_diagonal_component,
    get_expectation_value,
    get_fermion_number_operator,
    get_ground_state_rdm_from_qubit_op,
    get_polynomial_tensor,
    get_qubitop_from_coeffs_and_labels,
    get_qubitop_from_matrix,
    hf_rdm,
    remove_inactive_orbitals,
    reverse_qubit_order,
)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# isort: skip_file

# Out of order to fix circular import
from .bravyi_kitaev import bravyi_kitaev, inline_product, inline_sum
from .conversions import check_no_sympy, get_fermion_operator
from .jordan_wigner import jordan_wigner, jordan_wigner_one_body, jordan_wigner_two_body
from .reverse_jordan_wigner import reverse_jordan_wigner
from .term_reordering import (
    chemist_ordered,
    normal_ordered,
    normal_ordered_ladder_term,
    reorder,
)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Sequence, Union

import numpy as np

from orquestra.quantum.circuits import Circuit, GateOperation, Operation
from orquestra.quantum.circuits._circuit import split_circuit
from orquestra.quantum.circuits.layouts import CircuitConnectivity
from orquestra.quantum.openfermion import IsingOperator, QubitOperator, SymbolicOperator
from orquestra.quantum.wavefunction import Wavefunction

from ..distributions import (
    MeasurementOutcomeDistribution,
    create_bitstring_distribution_from_probability_distribution,
)
from ..measurements import ExpectationValues, Measurements, expectation_values_to_real
from ..openfermion import change_operator_type, get_expectation_value

# Note that in particular Wavefunction is a StateVector. However, for performance
# reasons QuantumSimulator uses numpy arrays internally.
StateVector = Union[Sequence[complex], np.ndarray]


class QuantumBackend(ABC):
    """Interface for implementing different quantum backends.

    Attributes:
        supports_batching: boolean flag indicating whether given backend
            supports batching circuits.
        batch_size: number of circuit runs in a single batch.
            If `supports_batching` is true should be a positive integer.
        number_of_circuits_run: number of circuits executed by this backend
        number_of_jobs_run: number of jobs executed by this backend. Will be different
            from `number_of_circuits_run` if batches are used.
    """

    supports_batching: bool = False
    batch_size: Optional[int] = None

    def __init__(self):
        self.number_of_circuits_run = 0
        self.number_of_jobs_run = 0

        if self.supports_batching:
            assert isinstance(self.batch_size, int)
            assert self.batch_size > 0

    @abstractmethod
    def run_circuit_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        """
        Method for executing the circuit and measuring the outcome.
        Args:
            circuit: quantum circuit to be executed.
            n_samples: The number of samples to collect.
        """
        if n_samples < 1:
            raise ValueError("Must sample given circuit at least once.")
        self.number_of_circuits_run += 1
        self.number_of_jobs_run += 1

        # NOTE: This value is only returned so that mypy doesn't complain.
        # You can remove this workaround when we reimplement counter increments in
        # a more type-elegant way.
        return Measurements()

    def run_circuitset_and_measure(
        self, circuits: Sequence[Circuit], n_samples: Sequence[int]
    ) -> List[Measurements]:
        """Run a set of circuits and measure a certain number of bitstrings.

        It may be useful to override this method for backends that support
        batching.

        Args:
            circuits: The circuits to execute.
            n_samples: The number of samples to collect for each circuit.
        """
        measurement_set: List[Measurements]

        if not self.supports_batching:
            measurement_set = []
            for circuit, n_samples_for_circuit in zip(circuits, n_samples):
                measurement_set.append(
                    self.run_circuit_and_measure(
                        circuit, n_samples=n_samples_for_circuit
                    )
                )

            return measurement_set
        else:
            self.number_of_circuits_run += len(circuits)
            if isinstance(self.batch_size, int):
                self.number_of_jobs_run += int(np.ceil(len(circuits) / self.batch_size))

            # This value is only returned so that mypy doesn't complain.
            # You can remove this workaround when we reimplement counter increments in
            # a more type-elegant way.
            measurement_set = []
            return measurement_set

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: int
    ) -> MeasurementOutcomeDistribution:
        """Calculates a measurement outcome distribution.

        Args:
            circuit: quantum circuit to be executed.

        Returns:
            Probability distribution of getting specific bistrings.

        """
        # Get the expectation values
        measurements = self.run_circuit_and_measure(circuit, n_samples)
        return measurements.get_distribution()


class QuantumSimulator(QuantumBackend):
    """Simulator capable of computing exact wavefunction.

    Note that in contrast to non-simulator QuantumBackends, simulators
    are capable of simulating operations that are not natively supported
    by libraries/services they wrap. Therefore, simulation of a circuit may
    get broken into several smaller circuits. Each native circuit run
    using the wrapped library or service counts towards number_of_circuits
    run and number_of_jobs_run. However, if simulated circuit comprises only
    natively supported operation AND concrete implementation does not change
    counting methodology, each simulated circuit corresponds to an increase
    of both those numbers by one.
    """

    @abstractmethod
    def __init__(
        self,
        noise_model: Optional[Any] = None,
        device_connectivity: Optional[CircuitConnectivity] = None,
    ):
        super().__init__()
        self.noise_model = noise_model
        self.device_connectivity = device_connectivity

    @abstractmethod
    def _get_wavefunction_from_native_circuit(
        self, circuit: Circuit, initial_state: StateVector
    ) -> StateVector:
        """Get wavefunction from circuit comprising only natively-supported operations.

        Args:
            circuit: circuit to simulate. Implementers of this function might assume
              that this circuit comprises only natively-supported operations as decided
              by self._is_supported predicate.
            initial_state: amplitudes of the initial state.
        Returns:
            StateVector representing amplitudes of final circuit state.
        """

    def is_natively_supported(self, operation: Operation) -> bool:
        """Determine if given operation is natively supported by this Simulator.

        This method can be as a predicate in split_circuit function.
        """
        return isinstance(operation, GateOperation)

    def get_wavefunction(
        self, circuit: Circuit, initial_state: Optional[StateVector] = None
    ) -> Wavefunction:
        """Returns a wavefunction representing quantum state produced by a circuit

        Args:
            circuit: quantum circuit to be executed.
            initial_state: a state from which the simulation starts.
              If not provided, the default |0...0> is used.
        """
        state: StateVector
        if initial_state is None:
            state = np.zeros(2**circuit.n_qubits)
            state[0] = 1
        else:
            state = initial_state

        for is_supported, subcircuit in split_circuit(
            circuit, self.is_natively_supported
        ):
            # Native subcircuits are passed through to the underlying simulator.
            # They also count towards number of circuits and number of jobs run.
            if is_supported:
                self.number_of_circuits_run += 1
                self.number_of_jobs_run += 1
                state = self._get_wavefunction_from_native_circuit(subcircuit, state)
            else:
                for operation in subcircuit.operations:
                    state = operation.apply(state)

        return Wavefunction(state)

    def get_exact_expectation_values(
        self, circuit: Circuit, operator: SymbolicOperator
    ) -> ExpectationValues:
        """Calculates the expectation values for given operator, based on the exact
        quantum state produced by circuit.

        Args:
            circuit: quantum circuit to be executed.
            operator: Operator for which we calculate the expectation value.
        """
        wavefunction = self.get_wavefunction(circuit)
        if isinstance(operator, IsingOperator):
            operator = change_operator_type(operator, QubitOperator)
        expectation_values = ExpectationValues(
            np.array([get_expectation_value(term, wavefunction) for term in operator])
        )
        expectation_values = expectation_values_to_real(expectation_values)
        return expectation_values

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: Optional[int] = None
    ) -> MeasurementOutcomeDistribution:
        """Calculates a bitstring distribution.

        Args:
            circuit: quantum circuit to be executed.

        Returns:
            Probability distribution of getting specific bistrings.
        """
        if n_samples is None:
            wavefunction = self.get_wavefunction(circuit)
            return create_bitstring_distribution_from_probability_distribution(
                wavefunction.get_probabilities()
            )
        else:
            # Get the expectation values
            measurements = self.run_circuit_and_measure(circuit, n_samples)
            return measurements.get_distribution()



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os

import numpy

from orquestra.quantum.openfermion.chem import MolecularData
from orquestra.quantum.openfermion.chem.reduced_hamiltonian import (
    make_reduced_hamiltonian,
)
from orquestra.quantum.openfermion.config import DATA_DIRECTORY
from orquestra.quantum.openfermion.linalg.sparse_tools import (
    get_number_preserving_sparse_operator,
)
from orquestra.quantum.openfermion.ops.representations import InteractionOperator
from orquestra.quantum.openfermion.transforms.opconversions import get_fermion_operator


def test_mrd_return_type():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(
        molecule.get_molecular_hamiltonian(), molecule.n_electrons
    )

    assert isinstance(reduced_ham, InteractionOperator)


def test_constant_one_body():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(
        molecule.get_molecular_hamiltonian(), molecule.n_electrons
    )

    assert numpy.isclose(reduced_ham.constant, molecule.nuclear_repulsion)
    assert numpy.allclose(reduced_ham.one_body_tensor, 0)


def test_fci_energy():
    filename = os.path.join(DATA_DIRECTORY, "H2_sto-3g_singlet_0.7414.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(
        molecule.get_molecular_hamiltonian(), molecule.n_electrons
    )
    numpy_ham = get_number_preserving_sparse_operator(
        get_fermion_operator(reduced_ham),
        molecule.n_qubits,
        num_electrons=molecule.n_electrons,
        spin_preserving=True,
    )

    w, _ = numpy.linalg.eigh(numpy_ham.toarray())
    assert numpy.isclose(molecule.fci_energy, w[0])

    filename = os.path.join(DATA_DIRECTORY, "H1-Li1_sto-3g_singlet_1.45.hdf5")
    molecule = MolecularData(filename=filename)
    reduced_ham = make_reduced_hamiltonian(
        molecule.get_molecular_hamiltonian(), molecule.n_electrons
    )
    numpy_ham = get_number_preserving_sparse_operator(
        get_fermion_operator(reduced_ham),
        molecule.n_qubits,
        num_electrons=molecule.n_electrons,
        spin_preserving=True,
    )

    w, _ = numpy.linalg.eigh(numpy_ham.toarray())
    assert numpy.isclose(molecule.fci_energy, w[0])



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for multiband Hubbard model module."""

import random
from enum import Enum

from orquestra.quantum.openfermion.hamiltonians.general_hubbard import number_operator
from orquestra.quantum.openfermion.ops import FermionOperator


class SpinPairs(Enum):
    """The spin orbitals corresponding to a pair of spatial orbitals."""

    SAME = 1
    ALL = 0
    DIFF = -1


def random_parameters(lattice, probability=0.5, distinguish_edges=False):
    parameters = {}
    edge_types = (
        ("onsite", "horizontal_neighbor", "vertical_neighbor")
        if distinguish_edges
        else ("onsite", "neighbor")
    )

    parameters["tunneling_parameters"] = [
        (edge_type, dofs, random.uniform(-1, 1))
        for edge_type in edge_types
        for dofs in lattice.dof_pairs_iter(edge_type == "onsite")
        if random.random() <= probability
    ]

    possible_spin_pairs = (
        (SpinPairs.ALL,) if lattice.spinless else (SpinPairs.SAME, SpinPairs.DIFF)
    )
    parameters["interaction_parameters"] = [
        (edge_type, dofs, random.uniform(-1, 1), spin_pairs)
        for edge_type in edge_types
        for spin_pairs in possible_spin_pairs
        for dofs in lattice.dof_pairs_iter(
            edge_type == "onsite" and spin_pairs in (SpinPairs.ALL, SpinPairs.SAME)
        )
        if random.random() <= probability
    ]

    parameters["potential_parameters"] = [
        (dof, random.uniform(-1, 1))
        for dof in lattice.dof_indices
        if random.random() <= probability
    ]

    if random.random() <= probability:
        parameters["magnetic_field"] = random.uniform(-1, 1)

    return parameters


def test_number_op():
    nop = FermionOperator(((0, 1), (0, 0)), coefficient=1.0)
    test_op = number_operator(0)
    assert test_op == nop

    test_op = number_operator(0, particle_hole_symmetry=True)
    assert test_op == nop - FermionOperator((), coefficient=0.5)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import random
import unittest

import numpy as np
import pkg_resources
import pytest

from orquestra.quantum.circuits import Circuit, X, Y, Z
from orquestra.quantum.measurements import ExpectationValues
from orquestra.quantum.openfermion.hamiltonians import fermi_hubbard
from orquestra.quantum.openfermion.linalg import (
    get_sparse_operator,
    jw_get_ground_state_at_particle_number,
    qubit_operator_sparse,
)
from orquestra.quantum.openfermion.ops import (
    FermionOperator,
    IsingOperator,
    QubitOperator,
)
from orquestra.quantum.openfermion.transforms import (
    get_fermion_operator,
    get_interaction_operator,
    jordan_wigner,
)
from orquestra.quantum.openfermion.zapata_utils._io import load_interaction_operator
from orquestra.quantum.utils import RNDSEED
from orquestra.quantum.wavefunction import Wavefunction


class TestQubitOperator(unittest.TestCase):
    def test_build_qubitoperator_from_coeffs_and_labels(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_qubitop_from_coeffs_and_labels,
        )

        # Given
        test_op = QubitOperator(((0, "Y"), (1, "X"), (2, "Z"), (4, "X")), 3.0j)
        coeffs = [3.0j]
        labels = [[2, 1, 3, 0, 1]]

        # When
        build_op = get_qubitop_from_coeffs_and_labels(coeffs, labels)

        # Then
        self.assertEqual(test_op, build_op)

    def test_qubitop_matrix_converion(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_qubitop_from_matrix,
        )

        # Given
        m = 4
        n = 2**m
        TOL = 10**-15
        random.seed(RNDSEED)
        A = np.array([[random.uniform(-1, 1) for x in range(n)] for y in range(n)])

        # When
        A_qubitop = get_qubitop_from_matrix(A)
        A_qubitop_matrix = np.array(qubit_operator_sparse(A_qubitop).todense())
        test_matrix = A_qubitop_matrix - A

        # Then
        for row in test_matrix:
            for elem in row:
                self.assertEqual(abs(elem) < TOL, True)

    def test_generate_random_qubitop(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            generate_random_qubitop,
        )

        # Given
        nqubits = 4
        nterms = 5
        nlocality = 2
        max_coeff = 1.5
        fixed_coeff = False

        # When
        qubit_op = generate_random_qubitop(
            nqubits, nterms, nlocality, max_coeff, fixed_coeff
        )
        # Then
        self.assertEqual(len(qubit_op.terms), nterms)
        for term, coefficient in qubit_op.terms.items():
            for i in range(nlocality):
                self.assertLess(term[i][0], nqubits)
            self.assertEqual(len(term), nlocality)
            self.assertLessEqual(np.abs(coefficient), max_coeff)

        # Given
        fixed_coeff = True
        # When
        qubit_op = generate_random_qubitop(
            nqubits, nterms, nlocality, max_coeff, fixed_coeff
        )
        # Then
        self.assertEqual(len(qubit_op.terms), nterms)
        for term, coefficient in qubit_op.terms.items():
            self.assertEqual(np.abs(coefficient), max_coeff)

    def test_evaluate_qubit_operator(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            evaluate_qubit_operator,
        )

        # Given
        qubit_op = QubitOperator("0.5 [] + 0.5 [Z1]")
        expectation_values = ExpectationValues([0.5, 0.5])
        # When
        value_estimate = evaluate_qubit_operator(qubit_op, expectation_values)
        # Then
        self.assertAlmostEqual(value_estimate, 0.5)

    def test_evaluate_qubit_operator_list(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            evaluate_qubit_operator_list,
        )

        # Given
        qubit_op_list = [
            QubitOperator("0.5 [] + 0.5 [Z1]"),
            QubitOperator("0.3 [X1] + 0.2[Y2]"),
        ]
        expectation_values = ExpectationValues([0.5, 0.5, 0.4, 0.6])
        # When
        value_estimate = evaluate_qubit_operator_list(qubit_op_list, expectation_values)
        # Then
        self.assertAlmostEqual(value_estimate, 0.74)

    def test_reverse_qubit_order(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            reverse_qubit_order,
        )

        # Given
        op1 = QubitOperator("[Z0 Z1]")
        op2 = QubitOperator("[Z1 Z0]")

        # When/Then
        self.assertEqual(op1, reverse_qubit_order(op2))

        # Given
        op1 = QubitOperator("Z0")
        op2 = QubitOperator("Z1")

        # When/Then
        self.assertEqual(op1, reverse_qubit_order(op2, n_qubits=2))
        self.assertEqual(op2, reverse_qubit_order(op1, n_qubits=2))

    def test_get_expectation_value(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_expectation_value,
        )

        """Check <Z0> and <Z1> for the state |100>"""
        # Given
        wf = Wavefunction([0, 1, 0, 0, 0, 0, 0, 0])
        op1 = QubitOperator("Z0")
        op2 = QubitOperator("Z1")
        # When
        exp_op1 = get_expectation_value(op1, wf, True)
        exp_op2 = get_expectation_value(op2, wf, True)

        # Then
        self.assertAlmostEqual(-1, exp_op1)
        self.assertAlmostEqual(1, exp_op2)

    def test_change_operator_type(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            change_operator_type,
        )

        # Given
        operator1 = QubitOperator("Z0 Z1", 4.5)
        operator2 = IsingOperator("Z0 Z1", 4.5)
        operator3 = IsingOperator()
        operator4 = IsingOperator("Z0", 0.5) + IsingOperator("Z1", 2.5)
        # When
        new_operator1 = change_operator_type(operator1, IsingOperator)
        new_operator2 = change_operator_type(operator2, QubitOperator)
        new_operator3 = change_operator_type(operator3, QubitOperator)
        new_operator4 = change_operator_type(operator4, QubitOperator)

        # Then
        self.assertEqual(IsingOperator("Z0 Z1", 4.5), new_operator1)
        self.assertEqual(QubitOperator("Z0 Z1", 4.5), new_operator2)
        self.assertEqual(QubitOperator(), new_operator3)
        self.assertEqual(
            QubitOperator("Z0", 0.5) + QubitOperator("Z1", 2.5), new_operator4
        )

    def test_get_fermion_number_operator(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_fermion_number_operator,
        )

        # Given
        n_qubits = 4
        n_particles = None
        correct_operator = get_interaction_operator(
            FermionOperator(
                """
        0.0 [] +
        1.0 [0^ 0] +
        1.0 [1^ 1] +
        1.0 [2^ 2] +
        1.0 [3^ 3]
        """
            )
        )

        # When
        number_operator = get_fermion_number_operator(n_qubits)

        # Then
        self.assertEqual(number_operator, correct_operator)

        # Given
        n_qubits = 4
        n_particles = 2
        correct_operator = get_interaction_operator(
            FermionOperator(
                """
        -2.0 [] +
        1.0 [0^ 0] +
        1.0 [1^ 1] +
        1.0 [2^ 2] +
        1.0 [3^ 3]
        """
            )
        )

        # When
        number_operator = get_fermion_number_operator(n_qubits, n_particles)

        # Then
        self.assertEqual(number_operator, correct_operator)

    def test_create_circuits_from_qubit_operator(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            create_circuits_from_qubit_operator,
        )

        # Initialize target
        circuit1 = Circuit([Z(0), X(1)])
        circuit2 = Circuit([Y(0), Z(1)])

        # Given
        qubit_op = QubitOperator("Z0 X1") + QubitOperator("Y0 Z1")

        # When
        pauli_circuits = create_circuits_from_qubit_operator(qubit_op)

        # Then
        self.assertEqual(pauli_circuits[0], circuit1)
        self.assertEqual(pauli_circuits[1], circuit2)


class TestOtherUtils(unittest.TestCase):
    def test_get_diagonal_component_polynomial_tensor(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_diagonal_component,
            get_polynomial_tensor,
        )

        fermion_op = FermionOperator("0^ 1^ 2^ 0 1 2", 1.0)
        fermion_op += FermionOperator("0^ 1^ 2^ 0 1 3", 2.0)
        fermion_op += FermionOperator((), 3.0)
        polynomial_tensor = get_polynomial_tensor(fermion_op)
        diagonal_op, remainder_op = get_diagonal_component(polynomial_tensor)
        self.assertTrue((diagonal_op + remainder_op) == polynomial_tensor)
        diagonal_qubit_op = jordan_wigner(get_fermion_operator(diagonal_op))
        remainder_qubit_op = jordan_wigner(get_fermion_operator(remainder_op))
        for term in diagonal_qubit_op.terms:
            for pauli in term:
                self.assertTrue(pauli[1] == "Z")
        for term in remainder_qubit_op.terms:
            is_diagonal = True
            for pauli in term:
                if pauli[1] != "Z":
                    is_diagonal = False
                    break
            self.assertFalse(is_diagonal)

    def test_get_diagonal_component_interaction_op(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_diagonal_component,
        )

        fermion_op = FermionOperator("1^ 1", 0.5)
        fermion_op += FermionOperator("2^ 2", 0.5)
        fermion_op += FermionOperator("1^ 2^ 0 3", 0.5)
        diagonal_op, remainder_op = get_diagonal_component(
            get_interaction_operator(fermion_op)
        )
        self.assertTrue(
            (diagonal_op + remainder_op) == get_interaction_operator(fermion_op)
        )
        diagonal_qubit_op = jordan_wigner(diagonal_op)
        remainder_qubit_op = jordan_wigner(remainder_op)
        for term in diagonal_qubit_op.terms:
            for pauli in term:
                self.assertTrue(pauli[1] == "Z")
        is_diagonal = True
        for term in remainder_qubit_op.terms:
            for pauli in term:
                if pauli[1] != "Z":
                    is_diagonal = False
                    break
        self.assertFalse(is_diagonal)

    def test_get_ground_state_rdm_from_qubit_op(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            get_ground_state_rdm_from_qubit_op,
        )

        # Given
        n_sites = 2
        U = 5.0
        fhm = fermi_hubbard(
            x_dimension=n_sites,
            y_dimension=1,
            tunneling=1.0,
            coulomb=U,
            chemical_potential=U / 2,
            magnetic_field=0,
            periodic=False,
            spinless=False,
            particle_hole_symmetry=False,
        )
        fhm_qubit = jordan_wigner(fhm)
        fhm_int = get_interaction_operator(fhm)
        e, wf = jw_get_ground_state_at_particle_number(
            get_sparse_operator(fhm), n_sites
        )

        # When
        rdm = get_ground_state_rdm_from_qubit_op(
            qubit_operator=fhm_qubit, n_particles=n_sites
        )

        # Then
        self.assertAlmostEqual(e, rdm.expectation(fhm_int))

    def test_remove_inactive_orbitals(self):
        from orquestra.quantum.openfermion.zapata_utils._utils import (
            hf_rdm,
            remove_inactive_orbitals,
        )

        fermion_ham = load_interaction_operator(
            pkg_resources.resource_filename(
                "orquestra.quantum.testing", "hamiltonian_HeH_plus_STO-3G.json"
            )
        )
        frozen_ham = remove_inactive_orbitals(fermion_ham, 1, 1)
        self.assertEqual(frozen_ham.one_body_tensor.shape[0], 2)

        hf_energy = hf_rdm(1, 1, 2).expectation(fermion_ham)
        self.assertAlmostEqual(frozen_ham.constant, hf_energy)


# Hamiltonians and energies from Psi4 H2 minimal basis
# first one is RHF, second one is H2- doublet with ROHF
@pytest.mark.parametrize(
    "hamiltonian, ref_energy, nalpha",
    [
        (
            load_interaction_operator(
                pkg_resources.resource_filename(
                    "orquestra.quantum.testing", "hamiltonian_H2_minimal_basis.json"
                )
            ),
            -0.8543376267387818,
            1,
        ),
        (
            load_interaction_operator(
                pkg_resources.resource_filename(
                    "orquestra.quantum.testing",
                    "hamiltonian_H2_minus_ROHF_minimal_basis.json",
                )
            ),
            -0.6857403043904364,
            2,
        ),
    ],
)
def test_hf_rdm_energy(hamiltonian, ref_energy, nalpha):
    from orquestra.quantum.openfermion.zapata_utils._utils import hf_rdm

    rdm = hf_rdm(nalpha, 1, 2)
    assert np.isclose(ref_energy, rdm.expectation(hamiltonian))



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import os
import unittest

import numpy as np

from orquestra.quantum.openfermion import (
    FermionOperator,
    InteractionRDM,
    IsingOperator,
    QubitOperator,
    get_interaction_operator,
    hermitian_conjugated,
)
from orquestra.quantum.openfermion.zapata_utils._io import (
    convert_dict_to_interaction_op,
    convert_dict_to_interaction_rdm,
    convert_dict_to_isingop,
    convert_dict_to_qubitop,
    convert_interaction_op_to_dict,
    convert_interaction_rdm_to_dict,
    convert_isingop_to_dict,
    convert_qubitop_to_dict,
    get_pauli_strings,
    load_interaction_operator,
    load_interaction_rdm,
    load_ising_operator,
    load_qubit_operator,
    load_qubit_operator_set,
    save_interaction_operator,
    save_interaction_rdm,
    save_ising_operator,
    save_qubit_operator,
    save_qubit_operator_set,
)
from orquestra.quantum.utils import convert_dict_to_array


class TestQubitOperator(unittest.TestCase):
    def setUp(self):
        n_modes = 2
        np.random.seed(0)
        one_body_tensor = np.random.rand(*(n_modes,) * 2)
        two_body_tensor = np.random.rand(*(n_modes,) * 4)
        self.interaction_rdm = InteractionRDM(one_body_tensor, two_body_tensor)

    def test_qubitop_to_dict_io(self):
        # Given
        qubit_op = QubitOperator(((0, "Y"), (1, "X"), (2, "Z"), (4, "X")), 3.0j)
        qubit_op += hermitian_conjugated(qubit_op)

        # When
        qubitop_dict = convert_qubitop_to_dict(qubit_op)
        recreated_qubit_op = convert_dict_to_qubitop(qubitop_dict)

        # Then
        self.assertEqual(recreated_qubit_op, qubit_op)

    def test_qubit_operator_io(self):
        # Given
        qubit_op = QubitOperator(((0, "Y"), (3, "X"), (8, "Z"), (11, "X")), 3.0j)

        # When
        save_qubit_operator(qubit_op, "qubit_op.json")
        loaded_op = load_qubit_operator("qubit_op.json")

        # Then
        self.assertEqual(qubit_op, loaded_op)

    def test_qubit_operator_set_io(self):
        qubit_op1 = QubitOperator(((0, "Y"), (3, "X"), (8, "Z"), (11, "X")), 3.0j)
        qubit_op2 = QubitOperator(((0, "Y"), (0, "X"), (7, "Z"), (14, "X")), 1.0j)

        qubit_operator_set = [qubit_op1, qubit_op2]
        save_qubit_operator_set(qubit_operator_set, "qubit_operator_set.json")
        loaded_qubit_operator_set = load_qubit_operator_set("qubit_operator_set.json")
        for i in range(len(qubit_operator_set)):
            self.assertEqual(qubit_operator_set[i], loaded_qubit_operator_set[i])
        os.remove("qubit_operator_set.json")

    def test_interaction_op_to_dict_io(self):
        # Given
        test_op = FermionOperator("1^ 2^ 3 4")
        test_op += hermitian_conjugated(test_op)
        interaction_op = get_interaction_operator(test_op)
        interaction_op.constant = 0.0

        # When
        interaction_op_dict = convert_interaction_op_to_dict(interaction_op)
        recreated_interaction_op = convert_dict_to_interaction_op(interaction_op_dict)

        # Then
        self.assertEqual(recreated_interaction_op, interaction_op)

    def test_interaction_operator_io(self):
        # Given
        test_op = FermionOperator("1^ 2^ 3 4")
        test_op += hermitian_conjugated(test_op)
        interaction_op = get_interaction_operator(test_op)
        interaction_op.constant = 0.0

        # When
        save_interaction_operator(interaction_op, "interaction_op.json")
        loaded_op = load_interaction_operator("interaction_op.json")

        # Then
        self.assertEqual(interaction_op, loaded_op)
        os.remove("interaction_op.json")

    def test_qubitop_io(self):
        # Given
        qubit_op = QubitOperator(((0, "Y"), (1, "X"), (2, "Z"), (4, "X")), 3.0j)

        # When
        save_qubit_operator(qubit_op, "qubit_op.json")
        loaded_op = load_qubit_operator("qubit_op.json")

        # Then
        self.assertEqual(qubit_op, loaded_op)
        os.remove("qubit_op.json")

    def test_get_pauli_strings(self):
        qubit_operator = (
            QubitOperator(((0, "X"), (1, "Y")))
            - 0.5 * QubitOperator(((1, "Y"),))
            + 0.5 * QubitOperator(())
        )
        constructed_list = get_pauli_strings(qubit_operator)
        target_list = ["X0Y1", "Y1", ""]
        self.assertListEqual(constructed_list, target_list)

    def test_isingop_to_dict_io(self):
        # Given
        ising_op = IsingOperator("[] + 3[Z0 Z1] + [Z1 Z2]")

        # When
        isingop_dict = convert_isingop_to_dict(ising_op)
        recreated_isingop = convert_dict_to_isingop(isingop_dict)

        # Then
        self.assertEqual(recreated_isingop, ising_op)

    def test_isingop_io(self):
        # Given
        ising_op = IsingOperator("[] + 3[Z0 Z1] + [Z1 Z2]")

        # When
        save_ising_operator(ising_op, "ising_op.json")
        loaded_op = load_ising_operator("ising_op.json")

        # Then
        self.assertEqual(ising_op, loaded_op)
        os.remove("ising_op.json")

    def test_interaction_rdm_io(self):
        # Given

        # When
        save_interaction_rdm(self.interaction_rdm, "interaction_rdm.json")
        loaded_interaction_rdm = load_interaction_rdm("interaction_rdm.json")

        # Then
        self.assertEqual(self.interaction_rdm, loaded_interaction_rdm)
        os.remove("interaction_rdm.json")

    def test_convert_interaction_rdm_to_dict(self):
        rdm_dict = convert_interaction_rdm_to_dict(self.interaction_rdm)

        self.assertTrue(
            np.allclose(
                convert_dict_to_array(rdm_dict["one_body_tensor"]),
                self.interaction_rdm.one_body_tensor,
            )
        )
        self.assertTrue(
            np.allclose(
                convert_dict_to_array(rdm_dict["two_body_tensor"]),
                self.interaction_rdm.two_body_tensor,
            )
        )

    def test_convert_dict_to_interaction_rdm(self):
        rdm_dict = convert_interaction_rdm_to_dict(self.interaction_rdm)
        converted_interaction_rdm = convert_dict_to_interaction_rdm(rdm_dict)

        self.assertEqual(self.interaction_rdm, converted_interaction_rdm)



import json
from collections import Counter
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import numpy as np

from orquestra.quantum.typing import AnyPath, LoadSource
from orquestra.quantum.utils import (
    convert_array_to_dict,
    convert_dict_to_array,
    ensure_open,
)


class Parities:
    """A class representing counts of parities for Pauli terms.

    Args:
        values (np.array): Number of observations of parities. See Attributes.
        correlations (list): Number of observations of pairwise products of terms.
            See Attributes.

    Attributes:
        values (np.array): an array of dimension N x 2 indicating how many times
            each Pauli term was observed with even and odd parity, where N is the
            number of Pauli terms. Here values[i][0] and values[i][1] correspond
            to the number of samples with even and odd parities for term P_i,
            respectively.
        correlations (list): a list of 3-dimensional numpy arrays indicating how
            many times each product of Pauli terms was observed with even and odd
            parity. Here correlations[i][j][k][0] and correlations[i][j][k][1]
            correspond to the number of samples with even and odd parities term P_j P_k
            in frame i, respectively.
    """

    def __init__(
        self, values: np.ndarray, correlations: Optional[List[np.ndarray]] = None
    ):
        self.values = values
        self.correlations = correlations

    def to_dict(self) -> dict:
        data: Dict[str, Any] = {"values": convert_array_to_dict(self.values)}
        if self.correlations:
            data["correlations"] = [
                convert_array_to_dict(arr) for arr in self.correlations
            ]
        return data

    @classmethod
    def from_dict(cls, data: dict):
        values = convert_dict_to_array(data["values"])
        if data.get("correlations"):
            correlations: Optional[List] = [
                convert_dict_to_array(arr) for arr in data["correlations"]
            ]
        else:
            correlations = None
        return cls(values, correlations)


def save_parities(parities: Parities, filename: AnyPath) -> None:
    """Save parities to a file.

    Args:
        parities (orquestra.quantum.measurement.Parities): the parities
        file (str or file-like object): the name of the file, or a file-like object
    """
    data = parities.to_dict()

    with open(filename, "w") as f:
        f.write(json.dumps(data, indent=2))


def load_parities(file: LoadSource) -> Parities:
    """Load parities from a file.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        orquestra.quantum.measurement.Parities: the parities
    """

    with ensure_open(file) as f:
        data = json.load(f)

    return Parities.from_dict(data)


def check_parity(
    bitstring: Union[str, Sequence[int]], marked_qubits: Iterable[int]
) -> bool:
    """Determine if the marked qubits have even parity for the given bitstring.

    Args:
        bitstring: The bitstring, either as a tuple or little endian string.
        marked_qubits: The qubits whose parity is to be determined.

    Returns:
        True if an even number of the marked qubits are in the 1 state, False
            otherwise.
    """
    result = True
    for qubit_index in marked_qubits:
        if bitstring[qubit_index] == "1" or bitstring[qubit_index] == 1:
            result = not result
    return result


def check_parity_of_vector(
    bitstrings_vector: np.ndarray, marked_qubits: Iterable[int]
) -> np.ndarray:
    """Determine if the marked qubits have even parity for each bitstring in the given
        vector.
        NOTE: This performs the same functionality as `check_parity` but is much
        faster as it uses vectorization to find the parity of multiple bitstrings
        at once.

    Args:
        bitstring: A 2d array of bitstrings whose size is number of bistrings * number
            of qubits
        marked_qubits: The qubits whose parity is to be determined.

    Returns:
        A 1d array with size equal to number of bitstrings. Each entry is 1 if an even
        number of the marked qubits of the corresponding bitstring are in the 1 state
        and 0 if otherwise.
    """
    if not marked_qubits:
        return np.ones(bitstrings_vector.shape[0])

    # Check if an even number of the marked qubits of each bitstring are in the 1 state
    bitstring_subset = bitstrings_vector[:, np.fromiter(marked_qubits, dtype=int)]
    return (bitstring_subset.sum(axis=1) + 1) % 2


def get_parities_from_measurements(
    measurements: List[Tuple[int]], ising_operator
) -> Parities:
    """Get expectation values from bitstrings.

    Args:
        measurements (list): the measured bitstrings
        ising_operator (openfermion.ops.IsingOperator): the operator

    Returns:
        orquestra.quantum.measurement.Parities: parities of each term in the operator
    """

    # leave this import inside this function to avoid circular imports
    # note that ising_operator in args is not typed because of this
    from orquestra.quantum.openfermion.ops import IsingOperator

    # check input format
    if not isinstance(ising_operator, IsingOperator):
        raise TypeError("Input operator not openfermion.IsingOperator")

    # Count number of occurrences of bitstrings
    bitstring_frequencies = Counter(measurements)
    bitstrings_vector = np.array([*bitstring_frequencies.keys()])
    bitstring_counts: np.ndarray = np.fromiter(
        bitstring_frequencies.values(), dtype=int
    )

    # Count parity occurrences
    values = []
    for _, term in enumerate(ising_operator.terms):
        marked_qubits = [op[0] for op in term]
        parity = check_parity_of_vector(bitstrings_vector, marked_qubits)

        true_parity_count: int = (parity * bitstring_counts).sum()
        false_parity_count: int = ((1 - parity) * bitstring_counts).sum()
        values.append([true_parity_count, false_parity_count])

    # Count parity occurrences for pairwise products of operators
    correlations = [np.zeros((len(ising_operator.terms), len(ising_operator.terms), 2))]
    for term1_index, term1 in enumerate(ising_operator.terms):
        for term2_index, term2 in enumerate(ising_operator.terms):
            marked_qubits_term1 = [op[0] for op in term1]
            marked_qubits_term2 = [op[0] for op in term2]

            parity1 = check_parity_of_vector(bitstrings_vector, marked_qubits_term1)
            parity2 = check_parity_of_vector(bitstrings_vector, marked_qubits_term2)

            equal_parities = np.abs(
                parity1 - parity2
            )  # 0 if parities are equal, 1 otherwise

            # Counts of bitstrings where parity is equal
            correlations[0][term1_index, term2_index][0] += (
                (1 - equal_parities) * bitstring_counts
            ).sum()

            # Counts of bitstrings where parity is not equal
            correlations[0][term1_index, term2_index][1] += (
                (equal_parities) * bitstring_counts
            ).sum()

    return Parities(np.array(values), correlations)



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
import random

import numpy as np

from orquestra.quantum.openfermion.ops import IsingOperator, QubitOperator
from orquestra.quantum.wavefunction import Wavefunction


def create_random_qubitop(nqubits, nterms, seed=None):
    """Generates random qubit operator acting on nqubits with nterms for testing
    purposes.

    The resulting qubit operator it saved to file in JSON format under 'qubitop.json'.

    Args:
        nqubits: integer
            The number of qubits in the qubit operator
        nterms: integer
            The number of terms in the qubit operator

        *** OPTIONAL ***
        seed: integer
            The see for the random number generator

    Returns:
        None, a Qubit Operator (openfermion.QubitOperator) object is saved under
            'qubitop.json'
    """
    NUM_QUBITS = range(0, nqubits)

    if seed is not None:
        random.seed(seed)

    # Initialize empty qubit operator
    qubitop = QubitOperator()

    # Loop over number of separate terms in qubit operator
    for i in range(0, nterms):
        # Choose number of paulis to measure in term
        num_paulis = random.choice(range(nqubits + 1))

        # Create empty list of qubits
        qubits = []

        # Create empty term
        full_term = ""

        # Loop over paulis
        for j in range(0, num_paulis):
            # Choose random qubit
            qubit_index = random.choice(NUM_QUBITS)
            while qubit_index in qubits:
                # Ensure qubit not already being measured in this term
                qubit_index = random.choice(NUM_QUBITS)
            qubits.append(qubit_index)

            # Choose pauli
            pauli_gate = random.choice(["X", "Y", "Z"])
            # Construct string
            full_term += pauli_gate + str(qubit_index) + " "
        # Add full term to qubit operator
        qubitop += QubitOperator(full_term)

    return qubitop


def create_random_isingop(nqubits, nterms, seed=None):
    """Generates random ising operator acting on nqubits with nterms for testing
        purposes.

    Args:
        nqubits: integer
            The number of qubits in the qubit operator
        nterms: integer
            The number of terms in the qubit operator

        *** OPTIONAL ***
        seed: integer
            The see for the random number generator

    Returns:
        an Ising Operator (openfermion.IsingOperator) object
    """
    NUM_QUBITS = range(0, nqubits)

    if seed is not None:
        random.seed(seed)

    # Initialize empty qubit operator
    isingop = IsingOperator()

    # Loop over number of separate terms in qubit operator
    for i in range(0, nterms):
        # Choose number of paulis to measure in term
        num_paulis = random.choice(range(nqubits + 1))

        # Create empty list of qubits
        qubits = []

        # Create empty term
        full_term = ""

        # Loop over paulis
        for j in range(0, num_paulis):
            # Choose random qubit
            qubit_index = random.choice(NUM_QUBITS)
            while qubit_index in qubits:
                # Ensure qubit not already being measured in this term
                qubit_index = random.choice(NUM_QUBITS)
            qubits.append(qubit_index)

            # Choose pauli
            pauli_gate = "Z"
            # Construct string
            full_term += pauli_gate + str(qubit_index) + " "
        # Add full term to qubit operator
        isingop += IsingOperator(full_term)

    return isingop


def create_random_wavefunction(n_qubits, seed=None):
    if seed:
        np.random.seed(seed)

    random_vector = [
        complex(a, b)
        for a, b in zip(np.random.rand(2**n_qubits), np.random.rand(2**n_qubits))
    ]
    normalization_factor = np.sqrt(np.sum(np.abs(random_vector) ** 2))
    random_vector /= normalization_factor

    return Wavefunction(random_vector)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# isort: skip_file


from .opconversions import (
    bravyi_kitaev,
    check_no_sympy,
    chemist_ordered,
    get_fermion_operator,
    jordan_wigner,
    jordan_wigner_one_body,
    jordan_wigner_two_body,
    normal_ordered,
    normal_ordered_ladder_term,
    reorder,
    reverse_jordan_wigner,
)
from .repconversions import (
    freeze_orbitals,
    get_interaction_operator,
    get_molecular_data,
    prune_unused_indices,
)



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests  jordan_wigner.py."""
import itertools
import os
import unittest

import numpy
import sympy

from orquestra.quantum.openfermion.chem import MolecularData
from orquestra.quantum.openfermion.config import DATA_DIRECTORY
from orquestra.quantum.openfermion.hamiltonians.special_operators import number_operator
from orquestra.quantum.openfermion.ops.operators import FermionOperator, QubitOperator
from orquestra.quantum.openfermion.ops.representations import InteractionOperator
from orquestra.quantum.openfermion.testing.testing_utils import (
    random_interaction_operator,
)
from orquestra.quantum.openfermion.transforms.opconversions import (
    get_fermion_operator,
    normal_ordered,
    reverse_jordan_wigner,
)
from orquestra.quantum.openfermion.transforms.opconversions.jordan_wigner import (
    _jordan_wigner_interaction_op,
    jordan_wigner,
    jordan_wigner_one_body,
    jordan_wigner_two_body,
)
from orquestra.quantum.openfermion.transforms.repconversions import (
    get_interaction_operator,
)
from orquestra.quantum.openfermion.utils import hermitian_conjugated


class JordanWignerTransformTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5

    def test_bad_input(self):
        with self.assertRaises(TypeError):
            jordan_wigner(3)

    def test_transm_raise3(self):
        raising = jordan_wigner(FermionOperator(((3, 1),)))
        self.assertEqual(len(raising.terms), 2)

        correct_operators_x = ((0, "Z"), (1, "Z"), (2, "Z"), (3, "X"))
        correct_operators_y = ((0, "Z"), (1, "Z"), (2, "Z"), (3, "Y"))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, -0.5j)

        self.assertEqual(raising.terms[correct_operators_x], 0.5)
        self.assertEqual(raising.terms[correct_operators_y], -0.5j)
        self.assertTrue(raising == qtermx + qtermy)

    def test_transm_hopping_sympy(self):
        coeff = sympy.Symbol("x")
        hopping = FermionOperator(((3, 1), (2, 0))) * coeff
        hopping += FermionOperator(((2, 1), (3, 0))) * coeff

        jw_hopping = jordan_wigner(hopping)
        print(jw_hopping)
        correct_operators_xx = ((2, "X"), (3, "X"))
        correct_operators_yy = ((2, "Y"), (3, "Y"))
        qtermxx = QubitOperator(correct_operators_xx, 0.5 * coeff)
        qtermyy = QubitOperator(correct_operators_yy, 0.5 * coeff)

        self.assertEqual(jw_hopping.terms[correct_operators_xx], 0.5 * coeff)
        self.assertEqual(jw_hopping.terms[correct_operators_yy], 0.5 * coeff)
        self.assertTrue(jw_hopping == qtermxx + qtermyy)

    def test_transm_raise3_sympy(self):
        coeff = sympy.Symbol("x")
        raising = jordan_wigner(coeff * FermionOperator(((3, 1),)))
        self.assertEqual(len(raising.terms), 2)

        correct_operators_x = ((0, "Z"), (1, "Z"), (2, "Z"), (3, "X"))
        correct_operators_y = ((0, "Z"), (1, "Z"), (2, "Z"), (3, "Y"))
        qtermx = QubitOperator(correct_operators_x, 0.5 * coeff)
        qtermy = QubitOperator(correct_operators_y, -0.5j * coeff)

        self.assertEqual(raising.terms[correct_operators_x], 0.5 * coeff)
        self.assertEqual(raising.terms[correct_operators_y], -0.5j * coeff)
        self.assertTrue(raising == qtermx + qtermy)

    def test_transm_raise1(self):
        raising = jordan_wigner(FermionOperator(((1, 1),)))

        correct_operators_x = ((0, "Z"), (1, "X"))
        correct_operators_y = ((0, "Z"), (1, "Y"))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, -0.5j)

        self.assertEqual(raising.terms[correct_operators_x], 0.5)
        self.assertEqual(raising.terms[correct_operators_y], -0.5j)
        self.assertTrue(raising == qtermx + qtermy)

    def test_transm_lower3(self):
        lowering = jordan_wigner(FermionOperator(((3, 0),)))

        correct_operators_x = ((0, "Z"), (1, "Z"), (2, "Z"), (3, "X"))
        correct_operators_y = ((0, "Z"), (1, "Z"), (2, "Z"), (3, "Y"))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_lower2(self):
        lowering = jordan_wigner(FermionOperator(((2, 0),)))

        correct_operators_x = ((0, "Z"), (1, "Z"), (2, "X"))
        correct_operators_y = ((0, "Z"), (1, "Z"), (2, "Y"))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_lower1(self):
        lowering = jordan_wigner(FermionOperator(((1, 0),)))

        correct_operators_x = ((0, "Z"), (1, "X"))
        correct_operators_y = ((0, "Z"), (1, "Y"))
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_lower0(self):
        lowering = jordan_wigner(FermionOperator(((0, 0),)))

        correct_operators_x = ((0, "X"),)
        correct_operators_y = ((0, "Y"),)
        qtermx = QubitOperator(correct_operators_x, 0.5)
        qtermy = QubitOperator(correct_operators_y, 0.5j)

        self.assertEqual(lowering.terms[correct_operators_x], 0.5)
        self.assertEqual(lowering.terms[correct_operators_y], 0.5j)
        self.assertTrue(lowering == qtermx + qtermy)

    def test_transm_raise3lower0(self):
        # recall that creation gets -1j on Y and annihilation gets +1j on Y.
        term = jordan_wigner(FermionOperator(((3, 1), (0, 0))))
        self.assertEqual(
            term.terms[((0, "X"), (1, "Z"), (2, "Z"), (3, "Y"))], 0.25 * 1 * -1j
        )
        self.assertEqual(
            term.terms[((0, "Y"), (1, "Z"), (2, "Z"), (3, "Y"))], 0.25 * 1j * -1j
        )
        self.assertEqual(
            term.terms[((0, "Y"), (1, "Z"), (2, "Z"), (3, "X"))], 0.25 * 1j * 1
        )
        self.assertEqual(
            term.terms[((0, "X"), (1, "Z"), (2, "Z"), (3, "X"))], 0.25 * 1 * 1
        )

    def test_transm_number(self):
        n = number_operator(self.n_qubits, 3)
        n_jw = jordan_wigner(n)
        self.assertEqual(n_jw.terms[((3, "Z"),)], -0.5)
        self.assertEqual(n_jw.terms[()], 0.5)
        self.assertEqual(len(n_jw.terms), 2)

    def test_ccr_offsite_even_ca(self):
        c2 = FermionOperator(((2, 1),))
        a4 = FermionOperator(((4, 0),))

        self.assertTrue(normal_ordered(c2 * a4) == normal_ordered(-a4 * c2))
        self.assertTrue(jordan_wigner(c2 * a4) == jordan_wigner(-a4 * c2))

    def test_ccr_offsite_odd_ca(self):
        c1 = FermionOperator(((1, 1),))
        a4 = FermionOperator(((4, 0),))
        self.assertTrue(normal_ordered(c1 * a4) == normal_ordered(-a4 * c1))

        self.assertTrue(jordan_wigner(c1 * a4) == jordan_wigner(-a4 * c1))

    def test_ccr_offsite_even_cc(self):
        c2 = FermionOperator(((2, 1),))
        c4 = FermionOperator(((4, 1),))
        self.assertTrue(normal_ordered(c2 * c4) == normal_ordered(-c4 * c2))

        self.assertTrue(jordan_wigner(c2 * c4) == jordan_wigner(-c4 * c2))

    def test_ccr_offsite_odd_cc(self):
        c1 = FermionOperator(((1, 1),))
        c4 = FermionOperator(((4, 1),))
        self.assertTrue(normal_ordered(c1 * c4) == normal_ordered(-c4 * c1))

        self.assertTrue(jordan_wigner(c1 * c4) == jordan_wigner(-c4 * c1))

    def test_ccr_offsite_even_aa(self):
        a2 = FermionOperator(((2, 0),))
        a4 = FermionOperator(((4, 0),))
        self.assertTrue(normal_ordered(a2 * a4) == normal_ordered(-a4 * a2))

        self.assertTrue(jordan_wigner(a2 * a4) == jordan_wigner(-a4 * a2))

    def test_ccr_offsite_odd_aa(self):
        a1 = FermionOperator(((1, 0),))
        a4 = FermionOperator(((4, 0),))
        self.assertTrue(normal_ordered(a1 * a4) == normal_ordered(-a4 * a1))

        self.assertTrue(jordan_wigner(a1 * a4) == jordan_wigner(-a4 * a1))

    def test_ccr_onsite(self):
        c1 = FermionOperator(((1, 1),))
        a1 = hermitian_conjugated(c1)
        self.assertTrue(
            normal_ordered(c1 * a1) == FermionOperator(()) - normal_ordered(a1 * c1)
        )
        self.assertTrue(
            jordan_wigner(c1 * a1) == QubitOperator(()) - jordan_wigner(a1 * c1)
        )

    def test_jordan_wigner_transm_op(self):
        n = number_operator(self.n_qubits)
        n_jw = jordan_wigner(n)
        self.assertEqual(self.n_qubits + 1, len(n_jw.terms))
        self.assertEqual(self.n_qubits / 2.0, n_jw.terms[()])
        for qubit in range(self.n_qubits):
            operators = ((qubit, "Z"),)
            self.assertEqual(n_jw.terms[operators], -0.5)


class InteractionOperatorsJWTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5
        self.constant = 0.0
        self.one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        self.two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float
        )
        self.interaction_operator = InteractionOperator(
            self.constant, self.one_body, self.two_body
        )

    def test_consistency(self):
        """Test consistency with JW for FermionOperators."""
        # Random interaction operator
        n_qubits = 5
        iop = random_interaction_operator(n_qubits, real=False)
        op1 = jordan_wigner(iop)
        op2 = jordan_wigner(get_fermion_operator(iop))

        self.assertEqual(op1, op2)

        # Interaction operator from molecule
        geometry = [("Li", (0.0, 0.0, 0.0)), ("H", (0.0, 0.0, 1.45))]
        basis = "sto-3g"
        multiplicity = 1

        filename = os.path.join(DATA_DIRECTORY, "H1-Li1_sto-3g_singlet_1.45")
        molecule = MolecularData(geometry, basis, multiplicity, filename=filename)
        molecule.load()

        iop = molecule.get_molecular_hamiltonian()
        op1 = jordan_wigner(iop)
        op2 = jordan_wigner(get_fermion_operator(iop))

        self.assertEqual(op1, op2)

    def testjordan_wigner_one_body(self):
        # Make sure it agrees with jordan_wigner(FermionTerm).
        for p in range(self.n_qubits):
            for q in range(self.n_qubits):
                coefficient = numpy.random.randn()
                if p != q:
                    coefficient += 1j * numpy.random.randn()
                # Get test qubit operator.
                test_operator = jordan_wigner_one_body(p, q, coefficient)

                # Get correct qubit operator.
                fermion_term = FermionOperator(((p, 1), (q, 0)), coefficient)
                if p != q:
                    fermion_term += FermionOperator(
                        ((q, 1), (p, 0)), coefficient.conjugate()
                    )
                correct_op = jordan_wigner(fermion_term)

                self.assertTrue(test_operator == correct_op)

    def testjordan_wigner_two_body(self):
        # Make sure it agrees with jordan_wigner(FermionTerm).
        for p, q, r, s in itertools.product(range(self.n_qubits), repeat=4):
            coefficient = numpy.random.randn()
            if set([p, q]) != set([r, s]):
                coefficient += 1j * numpy.random.randn()

            # Get test qubit operator.
            test_operator = jordan_wigner_two_body(p, q, r, s, coefficient)

            # Get correct qubit operator.
            fermion_term = FermionOperator(
                ((p, 1), (q, 1), (r, 0), (s, 0)), coefficient
            )
            if set([p, q]) != set([r, s]):
                fermion_term += FermionOperator(
                    ((s, 1), (r, 1), (q, 0), (p, 0)), coefficient.conjugate()
                )
            correct_op = jordan_wigner(fermion_term)

            self.assertTrue(
                test_operator == correct_op, str(test_operator - correct_op)
            )

    def test_jordan_wigner_twobody_interaction_op_allunique(self):
        test_op = FermionOperator("1^ 2^ 3 4")
        test_op += hermitian_conjugated(test_op)

        retransformed_test_op = reverse_jordan_wigner(
            jordan_wigner(get_interaction_operator(test_op))
        )

        self.assertTrue(
            normal_ordered(retransformed_test_op) == normal_ordered(test_op)
        )

    def test_jordan_wigner_twobody_interaction_op_reversal_symmetric(self):
        test_op = FermionOperator("1^ 2^ 2 1")
        test_op += hermitian_conjugated(test_op)
        self.assertTrue(
            jordan_wigner(test_op) == jordan_wigner(get_interaction_operator(test_op))
        )

    def test_jordan_wigner_interaction_op_too_few_n_qubits(self):
        with self.assertRaises(ValueError):
            _jordan_wigner_interaction_op(self.interaction_operator, self.n_qubits - 2)

    def test_jordan_wigner_interaction_op_with_zero_term(self):
        test_op = FermionOperator("1^ 2^ 3 4")
        test_op += hermitian_conjugated(test_op)

        interaction_op = get_interaction_operator(test_op)
        interaction_op.constant = 0.0

        retransformed_test_op = reverse_jordan_wigner(jordan_wigner(interaction_op))

        self.assertEqual(normal_ordered(retransformed_test_op), normal_ordered(test_op))


class GetInteractionOperatorTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5
        self.constant = 0.0
        self.one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        self.two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float
        )

    def test_get_interaction_operator_identity(self):
        interaction_operator = InteractionOperator(-2j, self.one_body, self.two_body)
        qubit_operator = jordan_wigner(interaction_operator)
        self.assertTrue(qubit_operator == -2j * QubitOperator(()))
        self.assertEqual(
            interaction_operator,
            get_interaction_operator(
                reverse_jordan_wigner(qubit_operator), self.n_qubits
            ),
        )

    def test_get_interaction_operator_one_body(self):
        interaction_operator = get_interaction_operator(
            FermionOperator("2^ 2"), self.n_qubits
        )
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        one_body[2, 2] = 1.0
        self.assertEqual(
            interaction_operator, InteractionOperator(0.0, one_body, self.two_body)
        )

    def test_get_interaction_operator_one_body_twoterm(self):
        interaction_operator = get_interaction_operator(
            FermionOperator("2^ 3", -2j) + FermionOperator("3^ 2", 3j), self.n_qubits
        )
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), complex)
        one_body[2, 3] = -2j
        one_body[3, 2] = 3j
        self.assertEqual(
            interaction_operator, InteractionOperator(0.0, one_body, self.two_body)
        )

    def test_get_interaction_operator_two_body(self):
        interaction_operator = get_interaction_operator(
            FermionOperator("2^ 2 3^ 4"), self.n_qubits
        )
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float
        )
        two_body[3, 2, 4, 2] = -1.0
        self.assertEqual(
            interaction_operator, InteractionOperator(0.0, self.one_body, two_body)
        )

    def test_get_interaction_operator_two_body_distinct(self):
        interaction_operator = get_interaction_operator(
            FermionOperator("0^ 1^ 2 3"), self.n_qubits
        )
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float
        )
        two_body[1, 0, 3, 2] = 1.0
        self.assertEqual(
            interaction_operator, InteractionOperator(0.0, self.one_body, two_body)
        )



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for commutators.py."""
import unittest

from orquestra.quantum.openfermion.ops.operators import FermionOperator, QubitOperator
from orquestra.quantum.openfermion.transforms import jordan_wigner
from orquestra.quantum.openfermion.transforms.opconversions import normal_ordered
from orquestra.quantum.openfermion.utils import hermitian_conjugated
from orquestra.quantum.openfermion.utils.commutators import (
    anticommutator,
    commutator,
    double_commutator,
)


class CommutatorTest(unittest.TestCase):
    def setUp(self):
        self.fermion_term = FermionOperator("1^ 2^ 3 4", -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term
        )
        self.qubit_operator = jordan_wigner(self.fermion_operator)

    def test_commutes_identity(self):
        com = commutator(FermionOperator.identity(), FermionOperator("2^ 3", 2.3))
        self.assertEqual(com, FermionOperator.zero())

    def test_commutes_no_intersection(self):
        com = commutator(FermionOperator("2^ 3"), FermionOperator("4^ 5^ 3"))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator.zero())

    def test_commutes_number_operators(self):
        com = commutator(FermionOperator("4^ 3^ 4 3"), FermionOperator("2^ 2"))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator.zero())

    def test_commutator_hopping_operators(self):
        com = commutator(3 * FermionOperator("1^ 2"), FermionOperator("2^ 3"))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator("1^ 3", 3))

    def test_commutator_hopping_with_single_number(self):
        com = commutator(FermionOperator("1^ 2", 1j), FermionOperator("1^ 1"))
        com = normal_ordered(com)
        self.assertEqual(com, -FermionOperator("1^ 2") * 1j)

    def test_commutator_hopping_with_double_number_one_intersection(self):
        com = commutator(FermionOperator("1^ 3"), FermionOperator("3^ 2^ 3 2"))
        com = normal_ordered(com)
        self.assertEqual(com, -FermionOperator("2^ 1^ 3 2"))

    def test_commutator_hopping_with_double_number_two_intersections(self):
        com = commutator(FermionOperator("2^ 3"), FermionOperator("3^ 2^ 3 2"))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator.zero())

    def test_commutator(self):
        operator_a = FermionOperator("")
        self.assertEqual(
            FermionOperator.zero(), commutator(operator_a, self.fermion_operator)
        )
        operator_b = QubitOperator("X1 Y2")
        self.assertEqual(
            commutator(self.qubit_operator, operator_b),
            (self.qubit_operator * operator_b - operator_b * self.qubit_operator),
        )

    def test_commutator_operator_a_bad_type(self):
        with self.assertRaises(TypeError):
            commutator(1, self.fermion_operator)

    def test_commutator_operator_b_bad_type(self):
        with self.assertRaises(TypeError):
            commutator(self.qubit_operator, "hello")

    def test_commutator_not_same_type(self):
        with self.assertRaises(TypeError):
            commutator(self.fermion_operator, self.qubit_operator)


class AnticommutatorTest(unittest.TestCase):
    def test_canonical_anticommutation_relations(self):
        op_1 = FermionOperator("3")
        op_1_dag = FermionOperator("3^")
        op_2 = FermionOperator("4")
        op_2_dag = FermionOperator("4^")
        zero = FermionOperator.zero()
        one = FermionOperator.identity()

        self.assertEqual(one, normal_ordered(anticommutator(op_1, op_1_dag)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1, op_2)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1, op_2_dag)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1_dag, op_2)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1_dag, op_2_dag)))
        self.assertEqual(one, normal_ordered(anticommutator(op_2, op_2_dag)))

    def test_anticommutator_not_same_type(self):
        with self.assertRaises(TypeError):
            anticommutator(FermionOperator(), QubitOperator())


class DoubleCommutatorTest(unittest.TestCase):
    def test_double_commutator_no_intersection_with_union_of_second_two(self):
        com = double_commutator(
            FermionOperator("4^ 3^ 6 5"),
            FermionOperator("2^ 1 0"),
            FermionOperator("0^"),
        )
        self.assertEqual(com, FermionOperator.zero())

    def test_double_commutator_more_info_not_hopping(self):
        com = double_commutator(
            FermionOperator("3^ 2"),
            FermionOperator("2^ 3") + FermionOperator("3^ 2"),
            FermionOperator("4^ 2^ 4 2"),
            indices2=set([2, 3]),
            indices3=set([2, 4]),
            is_hopping_operator2=True,
            is_hopping_operator3=False,
        )
        self.assertEqual(
            com, (FermionOperator("4^ 2^ 4 2") - FermionOperator("4^ 3^ 4 3"))
        )

    def test_double_commtator_more_info_both_hopping(self):
        com = double_commutator(
            FermionOperator("4^ 3^ 4 3"),
            FermionOperator("1^ 2", 2.1) + FermionOperator("2^ 1", 2.1),
            FermionOperator("1^ 3", -1.3) + FermionOperator("3^ 1", -1.3),
            indices2=set([1, 2]),
            indices3=set([1, 3]),
            is_hopping_operator2=True,
            is_hopping_operator3=True,
        )
        self.assertEqual(
            com,
            (FermionOperator("4^ 3^ 4 2", 2.73) + FermionOperator("4^ 2^ 4 3", 2.73)),
        )



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy

from orquestra.quantum.openfermion.chem import MolecularData
from orquestra.quantum.openfermion.config import EQ_TOLERANCE
from orquestra.quantum.openfermion.ops.operators import FermionOperator
from orquestra.quantum.openfermion.ops.representations import (
    InteractionOperator,
    InteractionOperatorError,
)
from orquestra.quantum.openfermion.transforms.opconversions import (
    check_no_sympy,
    normal_ordered,
)

# for breaking cyclic imports
from orquestra.quantum.openfermion.utils import operator_utils as op_utils


def get_interaction_operator(fermion_operator, n_qubits=None):
    r"""Convert a 2-body fermionic operator to InteractionOperator.

    This function should only be called on fermionic operators which
    consist of only a_p^\dagger a_q and a_p^\dagger a_q^\dagger a_r a_s
    terms. The one-body terms are stored in a matrix, one_body[p, q], and
    the two-body terms are stored in a tensor, two_body[p, q, r, s].

    Returns:
       interaction_operator: An instance of the InteractionOperator class.

    Raises:
        TypeError: Input must be a FermionOperator.
        TypeError: FermionOperator does not map to InteractionOperator.

    Warning:
        Even assuming that each creation or annihilation operator appears
        at most a constant number of times in the original operator, the
        runtime of this method is exponential in the number of qubits.
    """
    if not isinstance(fermion_operator, FermionOperator):
        raise TypeError("Input must be a FermionOperator.")

    check_no_sympy(fermion_operator)

    if n_qubits is None:
        n_qubits = op_utils.count_qubits(fermion_operator)
    if n_qubits < op_utils.count_qubits(fermion_operator):
        raise ValueError("Invalid number of qubits specified.")

    # Normal order the terms and initialize.
    fermion_operator = normal_ordered(fermion_operator)
    constant = 0.0
    one_body = numpy.zeros((n_qubits, n_qubits), complex)
    two_body = numpy.zeros((n_qubits, n_qubits, n_qubits, n_qubits), complex)

    # Loop through terms and assign to matrix.
    for term in fermion_operator.terms:
        coefficient = fermion_operator.terms[term]
        # Ignore this term if the coefficient is zero
        if abs(coefficient) < EQ_TOLERANCE:
            # not testable because normal_ordered kills
            # fermion terms lower than EQ_TOLERANCE
            continue  # pragma: no cover

        # Handle constant shift.
        if len(term) == 0:
            constant = coefficient

        elif len(term) == 2:
            # Handle one-body terms.
            if [operator[1] for operator in term] == [1, 0]:
                p, q = [operator[0] for operator in term]
                one_body[p, q] = coefficient
            else:
                raise InteractionOperatorError(
                    "FermionOperator does not map " "to InteractionOperator."
                )

        elif len(term) == 4:
            # Handle two-body terms.
            if [operator[1] for operator in term] == [1, 1, 0, 0]:
                p, q, r, s = [operator[0] for operator in term]
                two_body[p, q, r, s] = coefficient
            else:
                raise InteractionOperatorError(
                    "FermionOperator does not map " "to InteractionOperator."
                )

        else:
            # Handle non-molecular Hamiltonian.
            raise InteractionOperatorError(
                "FermionOperator does not map " "to InteractionOperator."
            )

    # Form InteractionOperator and return.
    interaction_operator = InteractionOperator(constant, one_body, two_body)
    return interaction_operator


def get_molecular_data(
    interaction_operator,
    geometry=None,
    basis=None,
    multiplicity=None,
    n_electrons=None,
    reduce_spin=True,
    data_directory=None,
):
    """Output a MolecularData object generated from an InteractionOperator

    Args:
        interaction_operator(InteractionOperator): two-body interaction
            operator defining the "molecular interaction" to be simulated.
        geometry(string or list of atoms):
        basis(string):  String denoting the basis set used to discretize the
            system.
        multiplicity(int): Spin multiplicity desired in the system.
        n_electrons(int): Number of electrons in the system
        reduce_spin(bool): True if one wishes to perform spin reduction on
            integrals that are given in interaction operator.  Assumes
            spatial (x) spin structure generically.

    Returns:
        molecule(MolecularData):
            Instance that captures the
            interaction_operator converted into the format that would come
            from an electronic structure package adorned with some meta-data
            that may be useful.
    """

    n_spin_orbitals = interaction_operator.n_qubits

    # Introduce bare molecular operator to fill
    molecule = MolecularData(
        geometry=geometry,
        basis=basis,
        multiplicity=multiplicity,
        data_directory=data_directory,
    )

    molecule.nuclear_repulsion = interaction_operator.constant

    # Remove spin from integrals and put into molecular operator
    if reduce_spin:
        reduction_indices = list(range(0, n_spin_orbitals, 2))
    else:
        reduction_indices = list(range(n_spin_orbitals))

    molecule.n_orbitals = len(reduction_indices)

    molecule.one_body_integrals = interaction_operator.one_body_tensor[
        numpy.ix_(reduction_indices, reduction_indices)
    ]
    molecule.two_body_integrals = interaction_operator.two_body_tensor[
        numpy.ix_(
            reduction_indices, reduction_indices, reduction_indices, reduction_indices
        )
    ]

    # Fill in other metadata
    molecule.overlap_integrals = numpy.eye(molecule.n_orbitals)
    molecule.n_qubits = n_spin_orbitals
    molecule.n_electrons = n_electrons
    molecule.multiplicity = multiplicity

    return molecule



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Commonly used operators (mainly instances of SymbolicOperator)."""
from typing import Optional

from orquestra.quantum.openfermion.ops.operators import FermionOperator


def number_operator(
    n_modes: int, mode: Optional[int] = None, coefficient=1.0, parity: int = -1
) -> FermionOperator:
    """Return a fermionic number operator.

    Args:
        n_modes (int): The number of modes in the system.
        mode (int, optional): The mode on which to return the number
            operator. If None, return total number operator on all sites.
        coefficient (float): The coefficient of the term.
        parity (int): Returns the fermionic number operator
                    if parity=-1 (default),
                    and returns the bosonic number operator
                    if parity=1.
    Returns:
        operator (FermionOperator)
    """

    if parity == -1:
        Op = FermionOperator
    elif parity == 1:
        # Op = BosonOperator
        raise ValueError("BosonOperator not available")
    else:
        raise ValueError("Invalid parity value: {}".format(parity))

    if mode is None:
        operator = Op()
        for m in range(n_modes):
            operator += number_operator(n_modes, m, coefficient, parity)
    else:
        operator = Op(((mode, 1), (mode, 0)), coefficient)
    return operator



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Test case prototypes that can be used in other projects.

Note that this file won't be executed on its own by pytest.
You need to define your own test cases that inherit from the ones defined here.

Note regarding testing specific gates.

To test that a gate is properly implemented, we can ask for its matrix representation
and check that each entry is correct. In some quantum simulator packages,
returning this matrix representation is either not possible or difficult to implement.
In such cases, we can check that the gate implementation is correct by ensuring that
the gate transforms input states to output states as expected. If the simulator has
the capability to provide the wavefunction as an output, then we can check that
the entries of the transformed wavefunction are correct. If the simulator does not
have the capability of providing the wavefunction as an output, but only gives
bitstring samples from the wavefunction, then we can check that the bitstring statistics
are as expected after taking sufficiently many samples. In both of these cases where
we cannot directly check the matrix corresponding to the gate, we must check the action
of the gate on multiple inputs (and outputs in the sampling case). We can picture
this process as a kind of "quantum process tomography" for gate unit testing.
Mathematically, correctness is ensured if the span of the input and outputs spans the
full vector space.  Checking a tomographically complete set of input and outputs could
be time consuming, especially in the case of sampling. Furthermore, we expect that the
bugs that will occur will lead to an effect on many inputs (rather than, say, a single
input-output pair).  Therefore, we are taking here a slightly lazy, but efficient
approach to testing these gates by testing how they transform a tomographically
incomplete set of input and outputs.

Gates tests use `backend_for_gates_test` instead of `backend` as an input parameter
because:
a) it has high chance of failing for noisy backends
b) having execution time in mind it's a good idea to use lower number of samples.
"""

import functools
from typing import List

import numpy as np
import pytest

from orquestra.quantum.api.estimation import EstimationTask
from orquestra.quantum.circuits import CNOT, Circuit, H, X, builtin_gate_by_name
from orquestra.quantum.openfermion import QubitOperator
from orquestra.quantum.wavefunction import Wavefunction

from ..distributions import MeasurementOutcomeDistribution
from ..estimation import estimate_expectation_values_by_averaging
from ..measurements import ExpectationValues, Measurements
from ..testing.test_cases_for_backend_tests import (
    one_qubit_non_parametric_gates_amplitudes_test_set,
    one_qubit_non_parametric_gates_exp_vals_test_set,
    one_qubit_parametric_gates_amplitudes_test_set,
    one_qubit_parametric_gates_exp_vals_test_set,
    two_qubit_non_parametric_gates_amplitudes_test_set,
    two_qubit_non_parametric_gates_exp_vals_test_set,
    two_qubit_parametric_gates_amplitudes_test_set,
    two_qubit_parametric_gates_exp_vals_test_set,
)


def skip_tests_for_excluded_gates(func):
    @functools.wraps(func)
    def _wrapper(self, **kwargs):
        tested_gate = kwargs["tested_gate"]
        if tested_gate not in self.gates_to_exclude:
            func(self, **kwargs)
        else:
            pytest.xfail(f"{tested_gate} gate is excluded for tests for this backend.")

    return _wrapper


class QuantumBackendTests:
    def test_run_circuit_and_measure_correct_indexing(self, backend):

        # Note: this test may fail with noisy devices
        # Given
        backend.number_of_circuits_run = 0
        backend.number_of_jobs_run = 0
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])
        n_samples = 100
        # When
        measurements = backend.run_circuit_and_measure(circuit, n_samples)

        # Then (since SPAM error could result in unexpected bitstrings, we make sure the
        # most common bitstring is the one we expect)
        counts = measurements.get_counts()
        assert max(counts, key=counts.get) == "001"
        assert backend.number_of_circuits_run == 1
        assert backend.number_of_jobs_run == 1

    @pytest.mark.parametrize("n_samples", [-1, 0])
    def test_run_circuit_and_measure_fails_for_invalid_n_samples(
        self, backend, n_samples
    ):
        # Given
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])

        # When
        with pytest.raises(ValueError):
            backend.run_circuit_and_measure(circuit, n_samples)

    @pytest.mark.parametrize("n_samples", [1, 2, 10, 100])
    def test_run_circuit_and_measure_correct_num_measurements_attribute(
        self, backend, n_samples
    ):
        # Given
        backend.number_of_circuits_run = 0
        backend.number_of_jobs_run = 0
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])

        # When
        measurements = backend.run_circuit_and_measure(circuit, n_samples)

        # Then
        assert isinstance(measurements, Measurements)
        assert len(measurements.bitstrings) == n_samples
        assert backend.number_of_circuits_run == 1
        assert backend.number_of_jobs_run == 1

    def test_if_all_measurements_have_the_same_number_of_bits(self, backend):
        # Given
        backend.number_of_circuits_run = 0
        backend.number_of_jobs_run = 0
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])

        # When
        measurements = backend.run_circuit_and_measure(circuit, n_samples=100)

        # Then
        assert all(len(bitstring) == 3 for bitstring in measurements.bitstrings)
        assert backend.number_of_circuits_run == 1
        assert backend.number_of_jobs_run == 1

    def test_run_circuitset_and_measure(self, backend):
        # Note: this test may fail with noisy devices
        # Given
        backend.number_of_circuits_run = 0
        backend.number_of_jobs_run = 0
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])
        n_samples = 100
        number_of_circuits = 25
        # When
        n_samples_per_circuit = [n_samples] * number_of_circuits
        measurements_set = backend.run_circuitset_and_measure(
            [circuit] * number_of_circuits, n_samples_per_circuit
        )

        # Then (since SPAM error could result in unexpected bitstrings, we make sure the
        # most common bitstring is the one we expect)
        for measurements in measurements_set:
            counts = measurements.get_counts()
            assert max(counts, key=counts.get) == "001"
        assert backend.number_of_circuits_run == number_of_circuits

        if backend.supports_batching:
            assert backend.number_of_jobs_run == int(
                np.ceil(number_of_circuits / backend.batch_size)
            )
        else:
            assert backend.number_of_jobs_run == number_of_circuits

    def test_run_circuitset_and_measure_n_samples(self, backend):
        # Note: this test may fail with noisy devices
        # Given
        backend.number_of_circuits_run = 0
        backend.number_of_jobs_run = 0
        first_circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])
        second_circuit = Circuit([X(0), X(1), X(2)])
        n_samples_per_circuit = [100, 105]

        # When
        measurements_set = backend.run_circuitset_and_measure(
            [first_circuit, second_circuit], n_samples_per_circuit
        )

        # Then (since SPAM error could result in unexpected bitstrings, we make sure the
        # most common bitstring is the one we expect)
        counts = measurements_set[0].get_counts()
        assert max(counts, key=counts.get) == "001"
        counts = measurements_set[1].get_counts()
        assert max(counts, key=counts.get) == "111"

        assert len(measurements_set[0].bitstrings) == n_samples_per_circuit[0]
        assert len(measurements_set[1].bitstrings) == n_samples_per_circuit[1]

        assert backend.number_of_circuits_run == 2


class QuantumBackendGatesTests:
    exp_val_spread = 1
    gates_to_exclude: List[str] = []

    @pytest.mark.parametrize(
        "initial_gate,tested_gate,target_values",
        one_qubit_non_parametric_gates_exp_vals_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_one_qubit_non_parametric_gates_using_expectation_values(
        self, backend_for_gates_test, initial_gate, tested_gate, target_values
    ):
        n_samples = 1000

        # Given
        gate_1 = builtin_gate_by_name(initial_gate)(0)
        gate_2 = builtin_gate_by_name(tested_gate)(0)

        circuit = Circuit([gate_1, gate_2])
        operators = [
            QubitOperator("[]"),
            QubitOperator("[X0]"),
            QubitOperator("[Y0]"),
            QubitOperator("[Z0]"),
        ]

        sigma = 1 / np.sqrt(n_samples)

        for i, operator in enumerate(operators):
            # When
            estimation_tasks = [EstimationTask(operator, circuit, n_samples)]
            expectation_values = estimate_expectation_values_by_averaging(
                backend_for_gates_test, estimation_tasks
            )
            calculated_value = expectation_values.values[0]

            # Then
            assert calculated_value == pytest.approx(
                target_values[i], abs=self.exp_val_spread * sigma * 3
            )

    @pytest.mark.parametrize(
        "initial_gate,tested_gate,params,target_values",
        one_qubit_parametric_gates_exp_vals_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_one_qubit_parametric_gates_using_expectation_values(
        self, backend_for_gates_test, initial_gate, tested_gate, params, target_values
    ):
        n_samples = 1000
        # Given
        gate_1 = builtin_gate_by_name(initial_gate)(0)
        gate_2 = builtin_gate_by_name(tested_gate)(*params)(0)

        circuit = Circuit([gate_1, gate_2])
        operators = [
            QubitOperator("[]"),
            QubitOperator("[X0]"),
            QubitOperator("[Y0]"),
            QubitOperator("[Z0]"),
        ]

        sigma = 1 / np.sqrt(n_samples)

        for i, operator in enumerate(operators):
            # When
            estimation_tasks = [EstimationTask(operator, circuit, n_samples)]
            expectation_values = estimate_expectation_values_by_averaging(
                backend_for_gates_test, estimation_tasks
            )
            calculated_value = expectation_values.values[0]

            # Then
            assert calculated_value == pytest.approx(
                target_values[i], abs=self.exp_val_spread * sigma * 3
            )

    @pytest.mark.parametrize(
        "initial_gates,tested_gate,operators,target_values",
        two_qubit_non_parametric_gates_exp_vals_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_two_qubit_non_parametric_gates_using_expectation_values(
        self,
        backend_for_gates_test,
        initial_gates,
        tested_gate,
        operators,
        target_values,
    ):
        n_samples = 1000

        # Given
        gate_1 = builtin_gate_by_name(initial_gates[0])(0)
        gate_2 = builtin_gate_by_name(initial_gates[1])(1)
        gate_3 = builtin_gate_by_name(tested_gate)(0, 1)

        circuit = Circuit([gate_1, gate_2, gate_3])

        sigma = 1 / np.sqrt(n_samples)

        for i, operator in enumerate(operators):
            # When
            operator = QubitOperator(operator)
            estimation_tasks = [EstimationTask(operator, circuit, n_samples)]
            expectation_values = estimate_expectation_values_by_averaging(
                backend_for_gates_test, estimation_tasks
            )
            calculated_value = expectation_values.values[0]

            # Then
            assert calculated_value == pytest.approx(
                target_values[i], abs=self.exp_val_spread * sigma * 5
            )

    @pytest.mark.parametrize(
        "initial_gates,tested_gate,params,operators,target_values",
        two_qubit_parametric_gates_exp_vals_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_two_qubit_parametric_gates_using_expectation_values(
        self,
        backend_for_gates_test,
        initial_gates,
        tested_gate,
        params,
        operators,
        target_values,
    ):
        n_samples = 1000

        # Given
        gate_1 = builtin_gate_by_name(initial_gates[0])(0)
        gate_2 = builtin_gate_by_name(initial_gates[1])(1)
        gate_3 = builtin_gate_by_name(tested_gate)(*params)(0, 1)

        circuit = Circuit([gate_1, gate_2, gate_3])

        sigma = 1 / np.sqrt(n_samples)

        for i, operator in enumerate(operators):
            # When
            operator = QubitOperator(operator)
            estimation_tasks = [EstimationTask(operator, circuit, n_samples)]
            expectation_values = estimate_expectation_values_by_averaging(
                backend_for_gates_test, estimation_tasks
            )
            calculated_value = expectation_values.values[0]

            # Then
            assert calculated_value == pytest.approx(
                target_values[i], abs=self.exp_val_spread * sigma * 5
            )


class QuantumSimulatorTests(QuantumBackendTests):
    atol_wavefunction = 1e-15
    atol_wavefunction_element = 1e-7
    atol_distribution = 1e-7

    def test_get_wavefunction(self, wf_simulator):
        # Given
        wf_simulator.number_of_circuits_run = 0
        wf_simulator.number_of_jobs_run = 0
        circuit = Circuit([H(0), CNOT(0, 1), CNOT(1, 2)])

        # When
        wavefunction = wf_simulator.get_wavefunction(circuit)

        # Then
        assert isinstance(wavefunction, Wavefunction)
        assert len(wavefunction.get_probabilities()) == 8
        assert wavefunction[0] == pytest.approx(
            (1 / np.sqrt(2) + 0j), abs=self.atol_wavefunction_element
        )
        assert wavefunction[7] == pytest.approx(
            (1 / np.sqrt(2) + 0j), abs=self.atol_wavefunction_element
        )
        assert wf_simulator.number_of_circuits_run == 1
        assert wf_simulator.number_of_jobs_run == 1

    def test_get_measurement_outcome_distribution_wf_simulators(self, wf_simulator):
        # Given
        wf_simulator.number_of_circuits_run = 0
        wf_simulator.number_of_jobs_run = 0
        circuit = Circuit([H(0), CNOT(0, 1), CNOT(1, 2)])

        # When
        distribution = wf_simulator.get_measurement_outcome_distribution(circuit)

        # Then
        assert isinstance(distribution, MeasurementOutcomeDistribution)
        assert distribution.get_number_of_subsystems() == 3
        assert distribution.distribution_dict[(0, 0, 0)] == pytest.approx(
            0.5, abs=self.atol_distribution
        )
        assert distribution.distribution_dict[(1, 1, 1)] == pytest.approx(
            0.5, abs=self.atol_distribution
        )
        assert wf_simulator.number_of_circuits_run == 1
        assert wf_simulator.number_of_jobs_run == 1

    def test_get_exact_expectation_values(self, wf_simulator):
        # Given
        wf_simulator.number_of_circuits_run = 0
        wf_simulator.number_of_jobs_run = 0
        circuit = Circuit([H(0), X(1)])
        operator = QubitOperator("[Z0] + 2[Z1]")
        target_expectation_values = ExpectationValues(np.array([0.0, -2.0]))

        # When
        expectation_values = wf_simulator.get_exact_expectation_values(
            circuit, operator
        )

        np.testing.assert_allclose(
            expectation_values.values,
            target_expectation_values.values,
            atol=self.atol_wavefunction,
        )
        assert wf_simulator.number_of_circuits_run == 1
        assert wf_simulator.number_of_jobs_run == 1

    def test_get_wavefunction_uses_provided_initial_state(self, wf_simulator):
        circuit = Circuit([H(0), H(1)])
        initial_state = np.array([0, 1, 0, 0])

        np.testing.assert_allclose(
            wf_simulator.get_wavefunction(circuit, initial_state=initial_state),
            np.array([0.5, -0.5, 0.5, -0.5]),
            atol=self.atol_wavefunction,
        )
        np.testing.assert_allclose(
            wf_simulator.get_wavefunction(circuit),
            0.5 * np.ones(4),
            atol=self.atol_wavefunction,
        )


class QuantumSimulatorGatesTest:
    atol_wavefunction = 1e-15
    gates_to_exclude: List[str] = []

    @pytest.mark.parametrize(
        "initial_gate,tested_gate,target_amplitudes",
        one_qubit_non_parametric_gates_amplitudes_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_one_qubit_non_parametric_gates_using_amplitudes(
        self,
        wf_simulator,
        initial_gate,
        tested_gate,
        target_amplitudes,
    ):
        # Given
        gate_1 = builtin_gate_by_name(initial_gate)(0)
        gate_2 = builtin_gate_by_name(tested_gate)(0)

        circuit = Circuit([gate_1, gate_2])

        # When
        wavefunction = wf_simulator.get_wavefunction(circuit)

        # Then
        np.testing.assert_allclose(
            wavefunction.amplitudes, target_amplitudes, atol=self.atol_wavefunction
        )

    @pytest.mark.parametrize(
        "initial_gate,tested_gate,params,target_amplitudes",
        one_qubit_parametric_gates_amplitudes_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_one_qubit_parametric_gates_using_amplitudes(
        self,
        wf_simulator,
        initial_gate,
        tested_gate,
        params,
        target_amplitudes,
    ):
        # Given
        gate_1 = builtin_gate_by_name(initial_gate)(0)
        gate_2 = builtin_gate_by_name(tested_gate)(*params)(0)

        circuit = Circuit([gate_1, gate_2])

        # When
        wavefunction = wf_simulator.get_wavefunction(circuit)

        # Then

        np.testing.assert_allclose(
            wavefunction.amplitudes, target_amplitudes, atol=self.atol_wavefunction
        )

    @pytest.mark.parametrize(
        "initial_gates,tested_gate,target_amplitudes",
        two_qubit_non_parametric_gates_amplitudes_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_two_qubit_non_parametric_gates_using_amplitudes(
        self,
        wf_simulator,
        initial_gates,
        tested_gate,
        target_amplitudes,
    ):
        # Given
        gate_1 = builtin_gate_by_name(initial_gates[0])(0)
        gate_2 = builtin_gate_by_name(initial_gates[1])(1)
        gate_3 = builtin_gate_by_name(tested_gate)(0, 1)

        circuit = Circuit([gate_1, gate_2, gate_3])

        # When
        wavefunction = wf_simulator.get_wavefunction(circuit)

        # Then
        np.testing.assert_allclose(
            wavefunction.amplitudes, target_amplitudes, atol=self.atol_wavefunction
        )

    @pytest.mark.parametrize(
        "initial_gates,tested_gate,params,target_amplitudes",
        two_qubit_parametric_gates_amplitudes_test_set,
    )
    @skip_tests_for_excluded_gates
    def test_two_qubit_parametric_gates_using_amplitudes(
        self,
        wf_simulator,
        initial_gates,
        tested_gate,
        params,
        target_amplitudes,
    ):
        # Given
        gate_1 = builtin_gate_by_name(initial_gates[0])(0)
        gate_2 = builtin_gate_by_name(initial_gates[1])(1)
        gate_3 = builtin_gate_by_name(tested_gate)(*params)(0, 1)

        circuit = Circuit([gate_1, gate_2, gate_3])
        # When
        wavefunction = wf_simulator.get_wavefunction(circuit)

        # Then
        np.testing.assert_allclose(
            wavefunction.amplitudes, target_amplitudes, atol=self.atol_wavefunction
        )



#   Copyright 2017 The OpenFermion Developers
#   Modifications copyright 2022 Zapata Computing, Inc. for compatibility reasons.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools
from typing import TYPE_CHECKING, Optional

from orquestra.quantum.openfermion import ops
from orquestra.quantum.openfermion.testing.testing_utils import (
    random_interaction_operator,
)

if TYPE_CHECKING:
    import orquestra.quantum.openfermion as openfermion


def random_interaction_operator_term(
    order: int,
    real: bool = True,
    seed: Optional[int] = None,
) -> "openfermion.InteractionOperator":
    """Generates a random interaction operator with non-zero coefficients only
    on terms corresponding to the given number of unique orbitals.

    The number of orbitals is equal to the given order.

    Args:
        order: How many unique orbitals the non-zero terms should correspond to.
        real: Whether or not the coefficients should be real. Defaults to True.
        seed: The seed. If None (default), uses np.random.
    """

    n_orbitals = order

    if order > 4:
        return ops.InteractionOperator.zero(order)

    operator = random_interaction_operator(n_orbitals, real=real, seed=seed)
    operator.constant = 0

    for indices in itertools.product(range(n_orbitals), repeat=2):
        if len(set(indices)) != order:
            operator.one_body_tensor[indices] = 0

    for indices in itertools.product(range(n_orbitals), repeat=4):
        if len(set(indices)) != order:
            operator.two_body_tensor[indices] = 0

    return operator



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Any, Dict, Optional

from sympy import Symbol

from orquestra.quantum.api.backend import QuantumSimulator, StateVector
from orquestra.quantum.circuits import Circuit, Operation
from orquestra.quantum.circuits.layouts import CircuitConnectivity
from orquestra.quantum.measurements import Measurements
from orquestra.quantum.wavefunction import sample_from_wavefunction


class SymbolicSimulator(QuantumSimulator):
    """A simulator computing wavefunction by consecutive gate matrix multiplication.

    Args:
        seed: the seed of the sampler
    """

    def __init__(
        self,
        noise_model: Optional[Any] = None,
        device_connectivity: Optional[CircuitConnectivity] = None,
        seed: Optional[int] = None,
    ):
        super().__init__(noise_model, device_connectivity)
        self._seed = seed

    def run_circuit_and_measure(
        self,
        circuit: Circuit,
        n_samples: int,
        symbol_map: Optional[Dict[Symbol, Any]] = None,
    ) -> Measurements:
        """Run a circuit and measure a certain number of bitstrings

        Args:
            circuit: the circuit to prepare the state
            n_samples: the number of bitstrings to sample
        """
        wavefunction = self.get_wavefunction(
            circuit if symbol_map is None else circuit.bind(symbol_map)
        )

        if wavefunction.free_symbols:
            raise ValueError(
                "Cannot sample from wavefunction with symbolic parameters."
            )

        bitstrings = sample_from_wavefunction(wavefunction, n_samples, self._seed)
        return Measurements(bitstrings)

    def _get_wavefunction_from_native_circuit(
        self, circuit: Circuit, initial_state: StateVector
    ) -> StateVector:
        state = initial_state

        for operation in circuit.operations:
            state = operation.apply(state)

        return state

    def is_natively_supported(self, operation: Operation) -> bool:
        return True
